/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen;

import java.io.IOException;
import java.io.Writer;

import com.ibm.jzos.ByteUtil;
import com.ibm.jzos.fields.AssemblerDatatypeFactory;
import com.ibm.jzos.fields.BinaryAsIntField;


/**
 * A object which maps a Common Header section of an Assembler or COBOL ADATA record. 
 * <p>
 * For more information, see:
 * <li>"HLASM Programmer's Guide: Associated Data File Output".</li>
 * <li>"Enterprise COBOL for z/OS Programming Guide"</li>
 */
public class ADataRecord {
	
	public static final int HEADER_LEN = 12;
	protected byte[] bytes;

	protected static AssemblerDatatypeFactory factory = new AssemblerDatatypeFactory();
	
	// ASMADATA DSECT                                                          
	// ADATA_LEN           DS    HL2     Variable length indicator
	//                     DS    XL2     Reserved                         @v4c
	//* Note: the previous four bytes are not actually in an ADATA record
	
	// ADATA_VERSION       DS    FL1     Product version number
	protected static BinaryAsIntField _ADATA_VERSION = (BinaryAsIntField)factory.getBinaryField(1,true);
	
	// ADATA_VERASM        EQU   16      High Level Assembler R1&2        @v4c
	public static final int ADATA_VERASM = 0x16;
	
	// ADATA_RECTYPE       DS    XL2     The record type                      X
	//                                    (See labels ADATA_RECnnn following)
	protected static BinaryAsIntField _ADATA_RECTYPE = (BinaryAsIntField)factory.getBinaryField(2,false);
	
	// ADATA_LEVEL         DS    FL1     The ASMADATA level
	protected static BinaryAsIntField _ADATA_LEVEL = (BinaryAsIntField)factory.getBinaryField(1, true);
	
	// ADATA_L1            EQU   1        - Associated DATA Level 1
	public static final int ADATA_L1 = 1;
	
	// ADATA_L2            EQU   2        - Associated DATA Level 2
	public static final int ADATA_L2 = 2;
	
	// ADATA_L3            EQU   3        - Associated DATA Level 3       @l1a 
	public static final int ADATA_L3 = 3;

	// ADATA_LVL           EQU   3        - Current level                 @l1a 
	public static final int ADATA_LVL = 3;

	// ADATA_CONT          DS    XL1     Continuation flag
	protected static BinaryAsIntField _ADATA_CONT = (BinaryAsIntField)factory.getBinaryField(1,false);
	
	// ADATA_NOTCONT       EQU   X'00'    - This record not continued
	public static final int ADATA_NOTCONT = 0x00;
	
	// ADATA_ISCONT        EQU   X'01'    - Continued on Next record
	public static final int ADATA_ISCONT = 0x01;
	
	// ADATA_EDITION       DS    FL1     Edition Number
	protected static BinaryAsIntField _ADATA_EDITION = (BinaryAsIntField)factory.getBinaryField(1, true);
	
	//                     DS    XL4     Reserved                         @v4c
	static { factory.getBinaryField(4,false); }
	
	// ADATA_DATA_LEN      DS    HL2     Associated data field length     @l1a
	protected static BinaryAsIntField _ADATA_DATA_LEN = (BinaryAsIntField)factory.getBinaryField(2,true);
	
	// ADATA_ORG           DS    0H      The length of the  record
	// ADATA_EQU_LEN       EQU   *-ASMADATA
	public static final int ADATA_EQU_LEN = factory.getOffset();
	// *                                                                  @v4a 
	
	
	/**
	 * Construct an instance on an existing byte array.
	 * @param buffer the byte array used as backing memory for the structure
	 */
	public ADataRecord(byte[] buffer) {
		setBytes(buffer);
		
		if (getSupportedEdition() > 0 && getEDITION() != getSupportedEdition() ) {
			String msg = "Expected record edition=" + getSupportedEdition()
						+ " but got edition=" + getEDITION();
			throw new IllegalArgumentException(msg);
		}
	}

	/**
	 * Answer the edition number that this (subtype instance) supports.
	 * The default is -1, having the special meaning that any edition is supported.
	 * Subclasses may override.
	 */
	public int getSupportedEdition() {
		return -1;  // means we don't care / support anything
	}
	
	/**
	 * Answer the underlying byte array mapped by this object.
	 * @return byte[]
	 */
	public byte[] getBytes() {
		return bytes;
	}
	
	/**
	 * Set the underlying buffer of bytes
	 * @param buffer the byte array
	 */
	public void setBytes(byte[] buffer) {
		if (buffer.length < HEADER_LEN) {
			throw new IllegalArgumentException("buffer length");
		}
		bytes = buffer;
	}
	
	public int getVERSION() {
		return _ADATA_VERSION.getInt(bytes);
	}
	
	public int getRECTYPE() {
		return _ADATA_RECTYPE.getInt(bytes);
	}
	
	public int getLEVEL() {
		return _ADATA_LEVEL.getInt(bytes);
	}

	public int getCONT() {
		return _ADATA_CONT.getInt(bytes);
	}
	
	public void setCONT(int value) {
		_ADATA_CONT.putInt(value, bytes);
	}
	
	public int getEDITION() {
		return _ADATA_EDITION.getInt(bytes);
	}
	
	public int getDATA_LEN() {
		return _ADATA_DATA_LEN.getInt(bytes);
	}
	
	public void setDATA_LEN(int value) {
		_ADATA_DATA_LEN.putInt(value, bytes);
	}
	
	/**
	 * Write a (debug) dump of the receiver
	 * @param wtr
	 * @throws IOException
	 */
	public void dumpOn(Writer wtr) throws IOException {
		String label = getClass().getName();
		int i = label.lastIndexOf('.');
		label = label.substring(i+1);
		
		ByteUtil.dumpHex(label, bytes,
				0, ADataRecord.ADATA_EQU_LEN + getDATA_LEN(), 32, 
				wtr);
	}
}

