/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen;

import java.io.BufferedInputStream;
import java.io.DataInputStream;
import java.io.EOFException;
import java.io.FileInputStream;
import java.io.IOException;

import com.ibm.jzos.ZFile;


/**
 * A class that reads ADataRecords from a file (or dataset) containing an ADATA file.
 * <p>
 * If the filename is prefixed with "//", it is treated as an MVS Dataset.  Otherwise
 * it is assumed to be a binary file (with RDWs) on a filesystem.
 * @see ZFile
 * @see ADataRecord
 */
public class ADataRecordReader {
	
	private String filename;
	private ZFile zfile;
	private byte[] zfilebuf;
	private DataInputStream dataInputStream;
	
	/**
	 * Construct an instance.
	 * @param filename the name of the ADATA file.  If prefixed with "//" an MVS
	 * dataset is assumed.  Otherwise a filesystem file is assumed.
	 */
	public ADataRecordReader(String filename) {
		this.filename = filename;
	}

	/**
	 * Reads the next ADataRecord.
	 * If the record is continued, multiple records are assembled together
	 * into a single record.
	 * <p>
	 * @return ADataRecord or null if the end of the file is reached
	 */
	public ADataRecord readRecord() throws IOException {
		
		byte[] bytes = null;
		try {
			bytes = readRecordBytes();
		} catch (EOFException ee) {
			cleanup();
			return null;
		}

		ADataRecord record = new ADataRecord(bytes);
		if (record.getCONT() == ADataRecord.ADATA_ISCONT) {
			// recursively combine continued records into one
			ADataRecord contRec = readRecord();
			combine(record, contRec);
		}

		return record;
	}
	
	/**
	 * Close the datasource; either a ZFile or DataInputStream
	 */
	private void cleanup() {
		try {
			if (dataInputStream != null) {
				dataInputStream.close();
			}
			if (zfile != null) {
				zfile.close();
			}
		} catch (IOException ioe) {
			//ignore
		}
	}
	
	/**
	 * Combine the contents of contRec into rec.
	 */
	private void combine(ADataRecord rec, ADataRecord contRec) {
		int newbuflen = ADataRecord.ADATA_EQU_LEN 
							+ rec.getDATA_LEN()
							+ contRec.getDATA_LEN();
		byte[] newbuf = new byte[newbuflen];
		System.arraycopy(rec.getBytes(), 0, 
							newbuf, 0, 
							ADataRecord.ADATA_EQU_LEN + rec.getDATA_LEN());
		System.arraycopy(contRec.getBytes(), ADataRecord.ADATA_EQU_LEN,
							newbuf, ADataRecord.ADATA_EQU_LEN + rec.getDATA_LEN(),
							contRec.getDATA_LEN());
		rec.setBytes(newbuf);
		rec.setCONT(ADataRecord.ADATA_NOTCONT);
	}

	/**
	 * Read the next set of record bytes from the datasource.
	 * @return byte[] the bytes
	 * @throws IOException if an error occurs
	 * @throws EOFException when eof occurs
	 */
	private byte[] readRecordBytes() throws IOException {
		byte[] bytes = null;
		if (dataInputStream == null && zfile == null) {
			if (filename.startsWith("//")) {
				zfile = new ZFile(filename, "rb,type=record,noseek");
				zfilebuf = new byte[zfile.getLrecl()];
			} else {
				dataInputStream = new DataInputStream(
									new BufferedInputStream(
										new FileInputStream(filename)));
			}
		}
		if (dataInputStream != null) {
			//Read IBM RDW: first two bytes are length
			int recLen = dataInputStream.readShort();
			//Second two bytes can be ignored
			dataInputStream.readShort();
			recLen -= 4; //Subtract the RDW bytes from overall length
			bytes = new byte[recLen];
			dataInputStream.readFully(bytes, 0, recLen);
		} else {
			int nRead;
			try {
				nRead = zfile.read(zfilebuf);
			} catch (IOException ioe) {
				cleanup();
				throw ioe;
			}
			if (nRead < 0) {
				// Create an EOFException to behave like DataInputStream EOF
				throw new EOFException(filename);
			} else {
				bytes = new byte[nRead];  // allocate only what we need
				System.arraycopy(zfilebuf, 0, bytes, 0, nRead);
			}
		}
		return bytes;
	}
	
}
