/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.asm;

import java.io.IOException;
import java.io.Writer;

import com.ibm.jzos.ByteUtil;
import com.ibm.jzos.fields.BinaryAsIntField;
import com.ibm.jzos.recordgen.ADataRecord;

/**
 * Maps a Adata DCDS record.
 * <p>
 * @see ADataRecord
 */
public class ADataDCDSRecord extends ADataRecord {
	
	public static final int HEADER_LEN = 0x0026; 
	
	static { factory.setOffset(ADataRecord.ADATA_EQU_LEN); }  // TODO hmmmm
	
	// ADATA_RECDCDS       EQU   X'0034'  - DC/DS
	public static final int ADATA_RECDCDS = 0x0034;
	
	// ADATA_EDNDCDS       EQU   X'01'    - DC/DS Edition
	public static final int ADATA_EDNDCDS = 0x01;
	
	// ADDCDS_ESDID        DS    FL4        ESDID for the source record
	protected static BinaryAsIntField _ADDCDS_ESDID = (BinaryAsIntField)factory.getBinaryField(4,true);
	
	// *                                                                  
	// ADDCDS_TYPE_DC      EQU   *,X'80'    Flag for DC
	public static final int ADDCDS_TYPE_DC = 0x80;
	public static final int ADDCDS_TYPE_DS = 0x00;
	
	// ADDCDS_TYPE_CXD     EQU   *,X'C0'    Flag for CXD within DC        
	public static final int ADDCDS_TYPE_CXD = 0xC0;

	// ADDCDS_TYPE_CCW     EQU   *,X'A0'    Flag for CCW within DC
	public static final int ADDCDS_TYPE_CCW = 0xA0;
	
	// ADDCDS_TYPE_DXD     EQU   *,X'40'    Flag for DXD
	public static final int ADDCDS_TYPE_DXD = 0x40;
	
	// ADDCDS_TEXT_REC     EQU   *,X'10'    X'0035' Text record exists
	public static final int ADDCDS_TEXT_REC = 0x10;
	
	// ADDCDS_TYPE_FLAG    DS    XL1        Type flag
	protected static BinaryAsIntField _ADDCDS_TYPE_FLAG = (BinaryAsIntField)factory.getBinaryField(1,false);
	
	// *                                                                  
	//                     DS    XL5        Reserved
	static { factory.getByteArrayField(5); }
	
	// ADDCDS_STMT         DS    FL4        Statement Number of source
	protected static BinaryAsIntField _ADDCDS_STMT = (BinaryAsIntField)factory.getBinaryField(4,true);
	
	// ADDCDS_OPS_NUM      DS    FL4        Number of operands            
	protected static BinaryAsIntField _ADDCDS_OPS_NUM = (BinaryAsIntField)factory.getBinaryField(4,true);

	// ADDCDS_OPS_OFF      DS    FL4        Offset of first operand     
	protected static BinaryAsIntField _ADDCDS_OPS_OFF = (BinaryAsIntField)factory.getBinaryField(4,true);
	
	// ADDCDS_EQU_LEN      EQU   *-ASMADATA                         

	/**
	 * Construct an instance on an existing byte array.
	 * @param buffer the byte array used as backing memory for the structure
	 */
	public ADataDCDSRecord(byte[] buffer) {
		super(buffer);
		if (buffer.length < HEADER_LEN + ADataRecord.HEADER_LEN) {
			throw new IllegalArgumentException("buffer length");
		}
		bytes = buffer;
	}
	
	/**
	 * Answer the edition number that this (subtype instance) supports.
	 */
	public int getSupportedEdition() {
		return 1;  
	}	

	public int getESDID() {
		return _ADDCDS_ESDID.getInt(bytes);
	}
	
	public int getTYPE_FLAG() {
		return _ADDCDS_TYPE_FLAG.getInt(bytes);
	}
	
	public int getOPS_NUM() {
		return _ADDCDS_OPS_NUM.getInt(bytes);
	}
	
	public int getOPS_OFF() {
		return _ADDCDS_OPS_OFF.getInt(bytes);
	}
	
	public int getSTMT() {
		return _ADDCDS_STMT.getInt(bytes);
	}
	
	public ADataDCDSOperand getFirstOperand() {
		if (getOPS_NUM() <= 0) {
			return null;
		}
		int off = getOPS_OFF();
		return new ADataDCDSOperand(bytes, off);
	}

	/**
	 * Write a (debug) dump of the receiver on a Writer
	 */
	public void dumpOn(Writer writer) throws IOException {
		//super.dumpOn(writer);
		writer.write("DCDS: TYPE_FLAG=0x");
		writer.write(ByteUtil.toHexString(getTYPE_FLAG(), 2));
		writer.write("  ESDID=");
		writer.write(Integer.toString(getESDID()));
		writer.write("  OPS_NUM=");
		writer.write(Integer.toString(getOPS_NUM()));
		writer.write(" \n");
		
		ADataDCDSOperand op = getFirstOperand();
		int opsNum = getOPS_NUM();
		for (int i=0; op != null && i<opsNum; i++) {
			op.dumpOn(writer);

			ADataDCDSValue value = op.getFirstValue();
			int valNum = op.getVALUES_NUM(); 
			for (int j=0; value != null && j<valNum; j++) {
				value.dumpOn(writer);
				value = value.getNextValue();
			}
			op = op.getNextOperand();
		}
	}
}
