/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.asm;

import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Iterator;

/**
 * An abstract superclass of objects that are components of a RecordSpec.
 * Each describes an item in an Assembler DSECT, and is used by 
 * {@link RecordClassGenerator} to generate a Java source class to describe
 * the Assembler DSECT.
 * <p>
 * @see RecordSpec
 * @see RecordClassGenerator
 * @see RecordSpecBuilder
 */
public abstract class ElementSpec {

	private ArrayList sourceLines = new ArrayList();
	private String label;
	private int sourceStmt = 0;

	/**
	 * Answer a List of Strings which are typically Assembler comments for the receiver's element.
	 * @return ArrayList
	 */
	public ArrayList getSourceLines() {
		return sourceLines;
	}

	/**
	 * @see #getSourceLines()
	 */
	public void setSourceLines(ArrayList sourceLines) {
		this.sourceLines = (ArrayList)sourceLines.clone();
	}

	/**
	 * Answer a String, possibly null, which is the Assembler label/symbol for the element
	 * @return String
	 */
	public String getLabel() {
		return label;
	}

	/**
	 * @see #getLabel()
	 */
	public void setLabel(String label) {
		this.label = label;
	}

	/**
	 * Answer an integer which is the Assembler source statement number
	 * @return int
	 */
	public int getSourceStmt() {
		return sourceStmt;
	}

	/**
	 * @see #getSourceStmt()
	 */
	public void setSourceStmt(int sourceStmt) {
		this.sourceStmt = sourceStmt;
	}

	/**
	 * Dump debug information for the receiver
	 * <p>
	 * @param writer
	 * @throws IOException
	 */
	public void dumpOn(Writer writer) throws IOException {
		
		writer.write("\n");
		for (Iterator i=getSourceLines().iterator(); i.hasNext();) {
			writer.write((String)i.next());
			writer.write("\n");
		}
		if (getLabel() == null) {
			writer.write("(null): ");
		} else {
			writer.write(getLabel());
			writer.write(": ");
		}
		writer.write(" stmt=" + getSourceStmt());
		writer.write("\n");
	}
	
	/**
	 * Static helper to get a simple class name for a Class
	 * @param c Class
	 * @return String
	 */
	static String getUnqualifiedClassname(Class c) {
		String fqn = c.getName();
		int i = fqn.lastIndexOf('.');
		if (i < 0) return fqn;
		return fqn.substring(i+1);
	}
}
