/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.asm;

import java.io.IOException;
import java.io.Writer;

import com.ibm.jzos.fields.BigDecimalAccessor;
import com.ibm.jzos.fields.BigIntegerAccessor;
import com.ibm.jzos.fields.Field;

/**
 * An object to describe a field in an Assembler DSECT.  Used by 
 * {@link RecordClassGenerator} to generate a Java source class to describe
 * the Assembler DSECT.  This object is created by {@link RecordSpecBuilder}.
 * <p>
 * @see RecordSpec
 * @see RecordClassGenerator
 * @see RecordSpecBuilder
 */
public class FieldSpec extends ElementSpec {

	private Field field;

	private int dupCount = 1;
	
	private int operandIndex = 0;
	
	private String typeAttr;
	private String typeExt;
	private int    bitOffset;
	
	/**
	 * Answer the dup (replication) count of the field.
	 * A value of zero means that subsequent fields overlay the current field.
	 * A value of one is a regular field, values greater than one is an "array" 
	 * @return int
	 */
	public int getDupCount() {
		return dupCount;
	}
	
	/**
	 * Answer true if the receiver is an array (dupCount>1) field
	 */
	public boolean isArray() {
		return getDupCount() > 1;
	}

	/**
	 * Answer whether this is for a BigDecimal or BigInteger field
	 */
	public boolean isBigDecimalOrInteger() {
		return field instanceof BigDecimalAccessor
			|| field instanceof BigIntegerAccessor;
	}
	/**
	 * @see #getDupCount()
	 */
	public void setDupCount(int dupCount) {
		this.dupCount = dupCount;
	}

	/**
	 * Answer a {@link com.ibm.jzos.fields.Field} subtype which describes the type mapping of the field.
	 */
	public Field getField() {
		return field;
	}

	/**
	 * @see #getField()
	 */
	public void setField(Field field) {
		this.field = field;
	}

	/**
	 * Answer the index of this FieldSpec within a list of FieldSpecs for the same assembler 
	 * statent and label
	 * @return int
	 */
	public int getOperandIndex() {
		return operandIndex;
	}

	/**
	 * @see #getOperandIndex()
	 */
	public void setOperandIndex(int operandIndex) {
		this.operandIndex = operandIndex;
	}

	public String getTypeAttr() {
		return typeAttr;
	}

	public void setTypeAttr(String typeAttr) {
		this.typeAttr = typeAttr;
	}

	public String getTypeExt() {
		return typeExt;
	}

	public void setTypeExt(String typeExt) {
		this.typeExt = typeExt;
	}

	public int getBitOffset() {
		return bitOffset;
	}

	public void setBitOffset(int bitOffset) {
		this.bitOffset = bitOffset;
	}

	/**
	 * Write a debug representation of the receiver.
	 */
	public void dumpOn(Writer writer) throws IOException {
		
		super.dumpOn(writer);
		
		writer.write(" dup=" + getDupCount());
		writer.write(getField().getClass().toString());
		writer.write("\n  offset=");
		writer.write(Integer.toString(getField().getOffset()));
		writer.write("  len=");
		writer.write(Integer.toString(getField().getByteLength()));
	}
	
	/**
	 * Answer true if the receiver's storage overlaps another FieldSpec
	 */
	public boolean overlaps(FieldSpec aFieldSpec) {
		if (this == aFieldSpec) return false;  // we don't overlap ourself
		Field first, second;
		if (field.getOffset() <= aFieldSpec.field.getOffset()) {
			first = field;
			second = aFieldSpec.field;
		} else {
			second = field;
			first = aFieldSpec.field;
		}
		// if either is length zero, then can't overlap
		if (first.getByteLength() == 0 || second.getByteLength() == 0) {
			return false;
		}
		int firstLimit = first.getOffset() + first.getByteLength();
		return firstLimit > second.getOffset();
	}

}
