/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.asm;

import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * A object which describes an abstract representation of the fields and 
 * constants in an Assembler DSECT.
 * <p>
 * @see RecordSpecBuilder
 */
public class RecordSpec {
	
	private ArrayList elementSpecs = new ArrayList();
	private String sectionName;
	
	/**
	 * Answer the name of the selected C/DSECT
	 */
	public String getSectionName() {
		return sectionName;
	}

	/**
	 * @see #getSectionName()
	 */
	public void setSectionName(String sectionName) {
		this.sectionName = sectionName;
	}

	/**
	 * Add a new ElementSpec (subtype) instance to the ordered
	 * list maintained by the receiver.
	 * @param elementSpec ElementSpec subtype instance
	 */
	public void addElementSpec(ElementSpec elementSpec) {
		elementSpecs.add(elementSpec);
	}

	/**
	 * Answer true if the receiver contains any fields that will require java.Math
	 */
	public boolean containsBigDecimalsOrIntegers() {
		for (Iterator i=elementSpecIterator(); i.hasNext();) {
			ElementSpec child = (ElementSpec)i.next();
			if (child instanceof FieldSpec) {
				FieldSpec fieldSpec = (FieldSpec)child;
				if (fieldSpec.isBigDecimalOrInteger()) return true;
			}
		}
		return false;
	}
	
	/**
	 * Answer an Iterator of the list of ElementSpec (subtypes instances)
	 * @return Iterator
	 */
	public Iterator elementSpecIterator() {
		return elementSpecs.iterator();
	}
	
	/**
	 * Answer an iterator over FieldSpecs that overlap the given FieldSpecs
	 */
	public Iterator overlappingFieldSpecsIterator(final FieldSpec fieldSpec) {
		return new Iterator() {
			private Iterator baseIterator = elementSpecIterator();
			private ElementSpec peek;
			private ElementSpec peek() {
				if (peek != null) return peek;
				if (baseIterator.hasNext()) {
					peek = (ElementSpec)baseIterator.next();
				}
				return peek;
			}
			public boolean hasNext() {
				while (peek() != null) {
					if (peek() instanceof FieldSpec
							&& fieldSpec.overlaps((FieldSpec)peek())) {
						return true;
					} else {
						peek = null; // discard/advance
					}
				}
				return false;
			}
			public Object next() {
				if (hasNext()) {
					Object answer = peek();
					peek = null; // consume it
					return answer;
				} else {
					throw new NoSuchElementException();
				}
			}
			public void remove() {
				throw new UnsupportedOperationException();
			}
			
		};
	}
	
	/**
	 * Write a debug representation of the Receiver
	 * @param writer Writer
	 * @throws IOException
	 */
	public void dumpOn(Writer writer) throws IOException {
		
		for (Iterator i=elementSpecIterator(); i.hasNext(); ) {
			ElementSpec elementSpec = (ElementSpec)i.next();
			elementSpec.dumpOn(writer);
		}
	}
		
 }
