/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2012. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.asm;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.StringReader;
import java.util.ArrayList;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import com.ibm.jzos.FileFactory;
import com.ibm.jzos.fields.Field;

/**
 * Parsed the XML description created by {@link RecordXMLGenerator) and
 * returns a RecordSpec describing an Assembler DSECT or CSECT.
 * <p/>
 * Note: This class is not intended to be used as a public API.
 */
public class RecordSpecXMLParser {

	private RecordSpec recordSpec = new RecordSpec();
	private XMLStreamReader reader;
	
	/**
	 * Parse an XML file
	 * @param filename the name of the XML file
	 * @return RecordSpec
	 * @throws IOException
	 */
	public RecordSpec process(String filename) throws IOException, XMLStreamException {

		reader = makeXMLReader(filename);
		String ename;
		
		while (reader.hasNext()) {
			switch (reader.next()) {
			case XMLStreamConstants.START_ELEMENT:
				ename = reader.getLocalName();
				if (ename.equalsIgnoreCase("symbol")) {
					processSymbol();
				} else if (ename.equalsIgnoreCase("field")) {
					processField();
				} else if (ename.equalsIgnoreCase("section")) {
					processSection();
				}
			}
		}
		return recordSpec;
	}

	protected void processSection() throws XMLStreamException {

		recordSpec.setSectionName(getAttributeValue("name"));
	}

	protected void processSymbol() throws XMLStreamException {

		ConstantSpec constantSpec = new ConstantSpec();
		constantSpec.setLabel(getAttributeValue("name"));
		constantSpec.setSourceStmt(getAttributeIntValue("sourceStmnt"));
		constantSpec.setRelocatable("true".equalsIgnoreCase(getAttributeValue("relocatable")));
		constantSpec.setTypeAttr(getAttributeValue("typeAttr"));
		constantSpec.setValue(new Integer(getAttributeIntValue("value")));
		
		constantSpec.setSourceLines(getSourceLines());
		recordSpec.addElementSpec(constantSpec);
	}
	
	protected void processField() throws XMLStreamException {
		
		FieldSpec fieldSpec = new FieldSpec();
		fieldSpec.setLabel(getAttributeValue("name"));
		fieldSpec.setSourceStmt(getAttributeIntValue("sourceStmnt"));
		int offset = getAttributeIntValue("offset");
		int byteLen = getAttributeIntValue("byteLength");
		fieldSpec.setOperandIndex(getAttributeIntValue("operandIndex"));
		char type = getAttributeValue("typeAttr", "X").charAt(0);
		char ext  = getAttributeValue("typeExt", " ").charAt(0);		
		fieldSpec.setDupCount(getAttributeIntValue("dupCount"));
		fieldSpec.setBitOffset(getAttributeIntValue("bitOffset"));
		if (fieldSpec.getBitOffset() == 0) {
			fieldSpec.setField(RecordSpecBuilder.makeField(type, ext, offset, byteLen)); 
		}
		fieldSpec.setTypeAttr(""+type);
		if (ext != ' ' && ext != '\0') 
			fieldSpec.setTypeExt(""+ext);
		
		fieldSpec.setSourceLines(getSourceLines());
		
		recordSpec.addElementSpec(fieldSpec);

	}
	
	private XMLStreamReader makeXMLReader(String filename) throws IOException, XMLStreamException {
		XMLStreamReader reader;
		XMLInputFactory xmlif = XMLInputFactory.newInstance();
		InputStream is = FileFactory.newInputStream(filename);
		reader = xmlif.createXMLStreamReader(is);
		return reader;
	}

	private String getAttributeValue(String name) {
		return reader.getAttributeValue(null, name);
	}

	private String getAttributeValue(String name, String def) {
		String v = reader.getAttributeValue(null, name);
		if (v == null) {
			return def;
		} else {
			return v;
		}
	}
	
	private int getAttributeIntValue(String name) {
		return getAttributeIntValue(name, 0);
		
	}
	private int getAttributeIntValue(String name, int def) {
		String v = reader.getAttributeValue(null, name);
		if (v == null) {
			return def;
		}
		return Integer.parseInt(v);
	}
	
	private ArrayList<String> getSourceLines() throws XMLStreamException {
		StringBuffer sb = new StringBuffer();
		boolean endElement = false;
		while (!endElement && reader.hasNext()) {
			switch (reader.next()) {
			case XMLStreamConstants.END_ELEMENT:
				endElement = true;
				break;
			case XMLStreamConstants.CHARACTERS:
				sb.append(new String(reader.getTextCharacters(),
										reader.getTextStart(),
										reader.getTextLength()));
				break;
			case XMLStreamConstants.CDATA:
				if (reader.hasText()) {
					sb.append(reader.getText());
				}
				break;
			}
		}
		ArrayList<String> lines = new ArrayList<String>();
		String line;
		boolean leading=true;
		BufferedReader sr = new BufferedReader(new StringReader(sb.toString()));
		try {
			while ((line = sr.readLine()) != null)	{
				if (leading && line.trim().length() == 0) {
					continue;
				} else {
					leading = false;
				}
				lines.add(line);
			}
		} catch (IOException canthappen) {
			throw new RuntimeException(canthappen);
		}
		// trim trailing empty lines
		for (int i=lines.size()-1; i>0; i--) {
			if (lines.get(i).trim().length() == 0) {
				lines.remove(i); 
			} else {
				break;
			}
		}
		return lines;
	}
}
