/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007, 2012. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.asm;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.StringWriter;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.StringTokenizer;

import javax.xml.stream.XMLOutputFactory;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

import com.ibm.jzos.fields.Field;
import com.ibm.jzos.recordgen.LibraryUtil;
import com.sun.xml.internal.txw2.output.IndentingXMLStreamWriter;

/**
 * Generates an XML description from an Assembler CSECT or DSECT.  The input to this
 * process is an ADATA file created by the IBM High-level Assembler.
 * The XML output is in a form that can be parsed by the {@link RecordClassGenerator}
 * and used to generate Java source code.
 * <p/>
 * The XML output closely parallels the ADATA file; for more insight as to meaning
 * of specific attributes please refer to "High Level Assembler: Programmer's Guide" /
 * "Appendix D: Associated data file output".
 * <p/>
 * The XML document structure is as follows:
 * <ul>
 * <li>element: section
 * 	<ul>
 * 	<li>attribute: name  (CSECT/DSECT name)</li>
 * 	</ul>
 * 	<ul>
 * 	<li>element: symbol (for each EQU symbol)
 *  	<ul>
 *  	<li>attribute: name</li>
 *  	<li>attribute: sourceStmnt (source line number)</li>
 *  	<li>attribute: typeAttr (symbol type)</li>
 *  	<li>attribute: relocatable (true if the symbol is relocatable; default=false)</li>
 *  	<li>attribute: value  (numeric or 'c' character value)</li>
 *  	</ul>
 *  </li>
 *  <li>element: field (for each DC or DS operand)
 *  	<ul>
 *  	<li>attribute: name</li>
 *  	<li>attribute: sourceStmnt (source line number)</li>
 *  	<li>attribute: offset (section offset to field)</li>
 *  	<li>attribute: byteLength (length of field)</li>
 *  	<li>attribute: operandIndex (index of this field operand in the same source statement)</li>
 *  	<li>attribute: typeAttr (operand type)</li>
 *  	<li>attribute: typeExt (operand type extension)</li>
 *  	<li>attribute: bitOffset (optional offset within byte for typeAttr=B)</li>
 *  	<li>attribute: dupCount (optional duplication factor; default=0)</li>
 *  	</ul>
 *  </li>
 * 	</ul>
 * </li>
 * </ul>
 * Section, symbol, and field elements may contain character data content consisting of the 
 * original assemmbler source line(s) associated with the element.
 * @see #main(String[]) main() method for information on running this class
 */
public class RecordXMLGenerator {

	private RecordSpec recordSpec;
	private XMLStreamWriter writer;
	private boolean includeSource = true;

	private static final String NL = System.getProperty("line.separator"); 
	private static final String INDENT = "   ";
	
	/**
	 * Main method / driver for generating XML descriptions from Assembler ADATA files.
	 * <p>
	 * Arguments consist of the following optional "name=value" pairs:  
	 * <ul>
	 * <li>adatafile=//MVS.DSN  (default: //DD:SYSADATA)</li>
	 * <li>outputFile=filename  (default: writes file to System.out (stdout))</li>
	 * <li>section=DSECTNAME    (default: process the first CSECT or DSECT found)</li>
	 * </ul>
	 * <p> 
	 * If the given <code>adatafile</code> is a pathname not starting with "//", then it is assumed
	 * to be the name of a binary file in the USS (or workstation) filesystem containing the 
	 * binary SYSADATA file.  In order to process SYSADATA files on the workstation, 
	 * they must be downloaded with IBM "RDWs" included.  This can be accomplished using FTP 
	 * along with the "bin" and "quote site rdw" commands".
	 * <p> 
	 * @param args key=value strings as described above.
	 * @throws IOException
	 */
	public static void main(String[] args) throws IOException {
		
		String adataFile = "//DD:SYSADATA";
		String outputFile = null;
		String section = null;
		boolean commentsPrecedeDeclaration = false;
		
		for (int i = 0; i<args.length; i++) {
			StringTokenizer strtok = new StringTokenizer(args[i],"=");
			if (strtok.countTokens() != 2) {
				throw new IllegalArgumentException("Argument '" + args[i] + "' not in the form: name=value");
			}
			String name = strtok.nextToken();
			String value = strtok.nextToken();
			if ("adatafile".equalsIgnoreCase(name)) {
				adataFile = value;
			} else if ("outputfile".equalsIgnoreCase(name)) {
				outputFile = value;
			} else if ("section".equalsIgnoreCase(name)) {
				section = value;
			} else if ("commentsPrecedeDeclaration".equalsIgnoreCase(name)) {
				commentsPrecedeDeclaration = "true".equalsIgnoreCase(value);				
			} else {
				throw new IllegalArgumentException("Argument '" + name + "' not recognized.");
			}
		}
		
		// Create a generator and configure it with the control arguments
		RecordXMLGenerator rxg = new RecordXMLGenerator();
		rxg.process(adataFile, section, commentsPrecedeDeclaration, outputFile);
	}
	
	/**
	 * Private ctor
	 * @see #main
	 */
	private RecordXMLGenerator() {
	}
	
	/**
	 * Construct an instance on a RecordSpec
	 */
	public RecordXMLGenerator(RecordSpec spec) {
		this.recordSpec = spec;
	}
	
	/**
	 * Process a single SYSADATA file and generate XML
	 */
	private void process(String adataFile, String sectionName, 
							boolean commentsPrecedeDeclaration, String xmlFile) throws IOException {
		
		// First, a RecordSpecBuilder converts the selected symbol of the ADATA
		// file into a RecordSpec model.
		adataFile = LibraryUtil.resolveLeadingTilde(adataFile);
		RecordSpecBuilder specBuilder = new RecordSpecBuilder();
		specBuilder.setSelectedSectionName(sectionName);
		specBuilder.setCommentsPrecedeDeclaration(commentsPrecedeDeclaration);
		recordSpec = specBuilder.process(adataFile);
		
		generateXmlOn(xmlFile);	
	}
	
	
	public void generateXmlOn(String xmlFile) throws IOException {

		OutputStream os = null;
		try {
			if (xmlFile != null) {
				File f = new File(LibraryUtil.resolveLeadingTilde(xmlFile));
				os = new FileOutputStream(f);
			} else {
				os = System.out;
			}
			generateXmlOn(os);		
		} finally {
			try { // to close the writer
				if (os != null) os.close(); 
			} catch(Throwable ignore) {}
		}
	}
	
	public void generateXmlOn(OutputStream os) throws IOException {
		
		try {
			writer = makeXmlWriter(os);
			startSectionElement();
			for (@SuppressWarnings("rawtypes")
			     Iterator itr=recordSpec.elementSpecIterator(); itr.hasNext();) {
				ElementSpec spec = (ElementSpec)itr.next();
				if (spec instanceof FieldSpec) {
					genFieldElement((FieldSpec)spec);
				} else if (spec instanceof ConstantSpec) {
					genSymbolElement((ConstantSpec)spec);
				}
			}
			endSectionElement();
			writer.flush();
			writer.close();
		} catch (XMLStreamException xse) {
			throw new IOException(xse);
		} finally {
			os.close();
		} 
	}
	
	protected XMLStreamWriter makeXmlWriter(OutputStream os) throws XMLStreamException {
		XMLOutputFactory xof = XMLOutputFactory.newInstance();
		XMLStreamWriter writer = xof.createXMLStreamWriter(os);
		
		return writer;
	}
	
	protected void startSectionElement() throws XMLStreamException {
		
		writer.writeStartDocument();
		writer.writeCharacters(NL);
		writer.writeStartElement("section");
		writer.writeAttribute("name", recordSpec.getSectionName());
		writer.writeCharacters(NL);
		
	}
	
	protected void endSectionElement() throws XMLStreamException {
		writer.writeEndElement();
		writer.writeCharacters(NL);
	}

	protected void genSymbolElement(ConstantSpec constantSpec) throws XMLStreamException {

		writer.writeCharacters(INDENT);
		writer.writeStartElement("symbol");
		writer.writeAttribute("name", constantSpec.getLabel());
		writer.writeAttribute("sourceStmnt", Integer.toString(constantSpec.getSourceStmt()));
		writer.writeAttribute("typeAttr", constantSpec.getTypeAttr());
		if (constantSpec.isRelocatable()) {
			writer.writeAttribute("relocatable", "true"); 			
		}
		writer.writeAttribute("value", constantSpec.getValue().toString());
		
		genSourceLines(constantSpec);
		
		writer.writeEndElement();
		writer.writeCharacters(NL);
	}
	
	protected void genFieldElement(FieldSpec fieldSpec) throws XMLStreamException {

		writer.writeCharacters(INDENT);
		writer.writeStartElement("field");
		if (fieldSpec.getLabel() != null) {
			writer.writeAttribute("name", fieldSpec.getLabel());
		}
		writer.writeAttribute("sourceStmnt", Integer.toString(fieldSpec.getSourceStmt()));

		Field field = fieldSpec.getField();

		writer.writeAttribute("offset", Integer.toString(field.getOffset()));
		writer.writeAttribute("byteLength", Integer.toString(field.getByteLength()));
		writer.writeAttribute("operandIndex", Integer.toString(fieldSpec.getOperandIndex()));
		writer.writeAttribute("typeAttr", fieldSpec.getTypeAttr());
		if (fieldSpec.getTypeExt() != null) {
			writer.writeAttribute("typeExt", fieldSpec.getTypeExt());			
		}
		if (fieldSpec.getBitOffset() != 0) {
			writer.writeAttribute("bitOffset", Integer.toString(fieldSpec.getBitOffset()));			
		}
		if (fieldSpec.getDupCount() != 0) {
			writer.writeAttribute("dupCount", Integer.toString(fieldSpec.getDupCount()));
		}
		
		if (fieldSpec.getOperandIndex() == 1) {
			genSourceLines(fieldSpec);
		}

		genSourceLines(fieldSpec);
		
		writer.writeEndElement();
		writer.writeCharacters(NL);
	}
	
	
	protected void genSourceLines(ElementSpec spec) throws XMLStreamException {
		
		if (!includeSource) return;
		
		@SuppressWarnings("unchecked")
		ArrayList<String> lines = (ArrayList<String>)spec.getSourceLines();
		if (lines.size() == 0) {
			return; 
		}
		// To avoid printing leading or trailing blank lines,
		// we find the first non-blank line
		int ifirst = 0;
		int ilast = lines.size() -1;
		for (ifirst=0; ifirst <= ilast; ifirst++) {
			String s = lines.get(ifirst);
			if (s.trim().length() > 0) {
				break;
			}
		}
		// Find the last non-blank line
		for (; ilast >= ifirst; ilast--) {
			String s = lines.get(ilast);
			if (s.trim().length() > 0) {
				break;
			}
		}		
		// Write the first through last non-blank lines to a String
		StringWriter sw = new StringWriter();
		sw.append(NL);  // leading NL
		for (int i=ifirst; i<= ilast; i++) {
			String s = lines.get(i);
			s = s.replaceAll("\\s+$", "");  // trim trailing white space
			sw.append(s);
			if (i < ilast) sw.append(NL);
		}
		writer.writeCharacters(sw.toString());
		writer.writeCharacters(NL);
		writer.writeCharacters(INDENT);
	}

	
}
