/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007, 2012. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.cobol;

import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamWriter;

import com.ibm.jzos.fields.*;

/**
 * An object to describe a field in a COBOL COPYBOOK.  Used by 
 * {@link RecordClassGenerator} to generate a Java source class to describe
 * the COPYBOOK.  This object is created by {@link RecordSpecTreeBuilder}.
 * <p>
 * @see RecordClassGenerator
 * @see RecordSpecTreeBuilder
 */
public class FieldSpec extends StorageSpec {

	private boolean containsOdoObject = false;
	private Field field;
	
	/** we have a copy of symRec.getPrecision() since Binary*Fields don't keep it for us */
	private int precision;
	
	/** we have a copy of the Comp5 setting since Binary*Fields don't remember */
	private boolean isComp5;

	static final String XML_ELEMENT_NAME = "field";
	
	/** 
	 * Answer true if this storage spec is or contains a field
	 * that is the object of a "OCCURS DEPENDS ON" clause.
	 * In other words, it contains the size of an ODO.
	 */
	public boolean containsOdoObject() {
		return containsOdoObject;
	}

	/**
	 * Answer a {@link com.ibm.jzos.fields.Field} subtype which describes the type mapping of the field.
	 * @return Field
	 */
	public Field getField() {
		return field;
	}

	/**
	 * Answers the number of digits of precision defined for the field.
	 * Kept here since some Fields don't have (such as BinaryFields).
	 */
	public int getPrecision() {
		return precision;
	}
	
	/**
	 * Answer true if the receiver is stored as an alpha-numeric field.
	 */
	public boolean isAlphaNumericField() {
		return
			field instanceof StringField
		||  field instanceof ExternalDecimalAsIntField
		||  field instanceof ExternalDecimalAsLongField
		||  field instanceof ExternalDecimalAsBigIntegerField
		||  field instanceof ExternalDecimalAsBigDecimalField
		||  field instanceof ExternalFloatField;
	}
	
	/**
	 * Answer whether the receiver is represented as an Array.
	 * Overridden to answer false if <code>isVarLenString</code> is true.
	 */
	public boolean isArray() {
		return super.isArray() && !isVarLenString();
	}
	
	/**
	 * Answer true if a (binary) field is Comp5
	 * @since 2.4.4
	 */
	public boolean isComp5() {
		return isComp5;
	}
	
	/**
	 * Return true if instance variable caches are used for accessor methods
	 * for this StorageSpec.  Overridden to force true if this is an
	 * ODO object
	 */
	public boolean isGenerateCachedAccessors() {
		return super.isGenerateCachedAccessors() || containsOdoObject();
	}
	
	/**
	 * Answer true if the receiver should be treated like a variable length
	 * String.  
	 */
	public boolean isVarLenString() {
		return getField() instanceof StringField
			&& getField().getByteLength() == 1
			&& isOdo();
	}
	
	/**
	 * Indicate whether the receiver IS an ODO object (the object
	 * of a OCCURS DEPENDING ON clause.
	 */
	public void setContainsOdoObject(boolean value) {
		containsOdoObject = value;
	}
	
	/**
	 * @see isComp5()
	 * @since 2.4.4
	 */
	public void setComp5(boolean value) {
		isComp5 = value;
	}
	
	/**
	 * @see #getField()
	 */
	public void setField(Field field) {
		this.field = field;
	}

	/**
	 * @see #getPrecision()
	 * @param precision
	 */
	public void setPrecision(int precision) {
		this.precision = precision;
	}

	/**
	 * A debug representation of the receiver.
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer(super.toString());
		sb.append(" ");
		String field = getField().getClass().toString();
		field = field.substring(field.lastIndexOf('.')+1);
		sb.append(field);
		return sb.toString();
	}
	
	/**
	 * @see ElementSpec#writeXmlOn(XMLStreamWriter, int)
	 */
	@Override
	public void writeXmlOn(XMLStreamWriter writer, int indentLevel)
			throws XMLStreamException {
		writeXmlNewline(writer, indentLevel);
		writer.writeStartElement(XML_ELEMENT_NAME);
		writeXmlAttributesOn(writer);
		writeXmlSourceLinesOn(writer);
		writeXmlChildrenOn(writer, indentLevel+1);
		writeXmlNewline(writer, indentLevel);
		writer.writeEndElement();
	}
	
	/**
	 * @see ElementSpec#writeXmlAttributesOn(XMLStreamWriter)
	 */
	@Override
	protected void writeXmlAttributesOn(XMLStreamWriter writer) throws XMLStreamException {
		super.writeXmlAttributesOn(writer);
		
		if (containsOdoObject()) {
			writer.writeAttribute("containsOdoObject", "true");
		}
	
		if (field instanceof StringField) {
			StringField f = (StringField)field;
			writer.writeAttribute("fieldType", "String");
			writer.writeAttribute("length", Integer.toString(f.getByteLength()));
			writer.writeAttribute("trim", f.isTrim()?"true":"false");
			writer.writeAttribute("padLeft", f.isPadLeft()?"true":"false");
			
		// The following three BinarAs* fields are the same, but must be individually
		// cast to get the precision :-(
		} else if (field instanceof BinaryAsIntField) {
			BinaryAsIntField f = (BinaryAsIntField)field;
			writer.writeAttribute("fieldType", "Binary");
			writer.writeAttribute("precision", Integer.toString(getPrecision()));
			writer.writeAttribute("signed", f.isSigned()?"true":"false");
			if (isComp5) writer.writeAttribute("comp5", "true");

		} else if (field instanceof BinaryAsLongField) {
			BinaryAsLongField f = (BinaryAsLongField)field;
			writer.writeAttribute("fieldType", "Binary");
			writer.writeAttribute("precision", Integer.toString(getPrecision()));
			writer.writeAttribute("signed", f.isSigned()?"true":"false");
			if (isComp5) writer.writeAttribute("comp5", "true");

		} else if (field instanceof BinaryAsBigDecimalField) {
			BinaryAsBigDecimalField f = (BinaryAsBigDecimalField)field;
			writer.writeAttribute("fieldType", "Binary");
			writer.writeAttribute("precision", Integer.toString(getPrecision()));
			writer.writeAttribute("signed", f.isSigned()?"true":"false");
			if (isComp5) writer.writeAttribute("comp5", "true");
			
			
		// The following four PackedDecimalAs* fields are the same, but must be individually
		// cast to get the attributes :-(
		} else if (field instanceof PackedDecimalAsBigDecimalField) {
			PackedDecimalAsBigDecimalField f = (PackedDecimalAsBigDecimalField)field;
			writer.writeAttribute("fieldType", "PackedDecimal");
			writer.writeAttribute("precision", Integer.toString(getPrecision()));
			writer.writeAttribute("scale", Integer.toString(f.getScale()));
			writer.writeAttribute("signed", f.isSigned()?"true":"false");
		} else if (field instanceof PackedDecimalAsBigIntegerField) {
			PackedDecimalAsBigIntegerField f = (PackedDecimalAsBigIntegerField)field;
			writer.writeAttribute("fieldType", "PackedDecimal");
			writer.writeAttribute("precision", Integer.toString(getPrecision()));
			writer.writeAttribute("scale", Integer.toString(f.getScale()));
			writer.writeAttribute("signed", f.isSigned()?"true":"false");
		} else if (field instanceof PackedDecimalAsLongField) {
			PackedDecimalAsLongField f = (PackedDecimalAsLongField)field;
			writer.writeAttribute("fieldType", "PackedDecimal");
			writer.writeAttribute("precision", Integer.toString(getPrecision()));
			writer.writeAttribute("scale", Integer.toString(0));
			writer.writeAttribute("signed", f.isSigned()?"true":"false");
		} else if (field instanceof PackedDecimalAsIntField) {
			PackedDecimalAsIntField f = (PackedDecimalAsIntField)field;
			writer.writeAttribute("fieldType", "PackedDecimal");
			writer.writeAttribute("precision", Integer.toString(getPrecision()));
			writer.writeAttribute("scale", Integer.toString(0));
			writer.writeAttribute("signed", f.isSigned()?"true":"false");

		// The following four ExternalDecimalAs* fields are the same, but must be individually
		// cast to get the attributes :-(
		} else if (field instanceof ExternalDecimalAsBigDecimalField) {
			ExternalDecimalAsBigDecimalField f = (ExternalDecimalAsBigDecimalField)field;
			writer.writeAttribute("fieldType", "ExternalDecimal");
			writer.writeAttribute("precision", Integer.toString(getPrecision()));
			writer.writeAttribute("scale", Integer.toString(f.getScale()));
			writer.writeAttribute("signed", f.isSigned()?"true":"false");
			if (f.isSigned()) {
				writer.writeAttribute("signTrailing", f.isSignTrailing()?"true":"false");
				writer.writeAttribute("signExternal", f.isSignExternal()?"true":"false");
			}
			writer.writeAttribute("blankWhenZero", f.isBlankWhenZero()?"true":"false");
		} else if (field instanceof ExternalDecimalAsBigIntegerField) {
			ExternalDecimalAsBigIntegerField f = (ExternalDecimalAsBigIntegerField)field;
			writer.writeAttribute("fieldType", "ExternalDecimal");
			writer.writeAttribute("precision", Integer.toString(getPrecision()));
			writer.writeAttribute("scale", Integer.toString(f.getScale()));
			writer.writeAttribute("signed", f.isSigned()?"true":"false");
			if (f.isSigned()) {
				writer.writeAttribute("signTrailing", f.isSignTrailing()?"true":"false");
				writer.writeAttribute("signExternal", f.isSignExternal()?"true":"false");
			}
			writer.writeAttribute("blankWhenZero", f.isBlankWhenZero()?"true":"false");
		} else if (field instanceof ExternalDecimalAsIntField) {
			ExternalDecimalAsIntField f = (ExternalDecimalAsIntField)field;
			writer.writeAttribute("fieldType", "ExternalDecimal");
			writer.writeAttribute("precision", Integer.toString(getPrecision()));
			writer.writeAttribute("scale", Integer.toString(0));
			writer.writeAttribute("signed", f.isSigned()?"true":"false");
			if (f.isSigned()) {
				writer.writeAttribute("signTrailing", f.isSignTrailing()?"true":"false");
				writer.writeAttribute("signExternal", f.isSignExternal()?"true":"false");
			}
			writer.writeAttribute("blankWhenZero", f.isBlankWhenZero()?"true":"false");
		} else if (field instanceof ExternalDecimalAsLongField) {
			ExternalDecimalAsLongField f = (ExternalDecimalAsLongField)field;
			writer.writeAttribute("fieldType", "ExternalDecimal");
			writer.writeAttribute("precision", Integer.toString(getPrecision()));
			writer.writeAttribute("scale", Integer.toString(0));
			writer.writeAttribute("signed", f.isSigned()?"true":"false");
			if (f.isSigned()) {
				writer.writeAttribute("signTrailing", f.isSignTrailing()?"true":"false");
				writer.writeAttribute("signExternal", f.isSignExternal()?"true":"false");
			}
			writer.writeAttribute("blankWhenZero", f.isBlankWhenZero()?"true":"false");

		} else if (field instanceof ExternalFloatField) {
			ExternalFloatField f = (ExternalFloatField)field;
			writer.writeAttribute("fieldType", "ExternalFloat");
			writer.writeAttribute("precision", Integer.toString(getPrecision()));
			writer.writeAttribute("scale", Integer.toString(f.getScale()));
			writer.writeAttribute("showMantissaPlusSign", f.isShowMantissaPlusSign()?"true":"false");
			writer.writeAttribute("showExponentPlusSign", f.isShowExponentPlusSign()?"true":"false");
			writer.writeAttribute("impliedDecimal", f.isImpliedDecimal()?"true":"false");			
			
		} else if (field instanceof IbmFloatField) {
			writer.writeAttribute("fieldType", "IbmFloat");
			
		} else if (field instanceof IbmDoubleField) {
			writer.writeAttribute("fieldType", "IbmDouble");

		} else if (field instanceof ByteArrayField) {
			writer.writeAttribute("fieldType", "ByteArray");
			writer.writeAttribute("length", Integer.toString(field.getByteLength()));
		}
			
	}	
	
	/**
	 * @see ElementSpec#parseXmlFrom(RecordSpecTreeXMLParser)
	 */
	@Override
	protected void parseXmlFrom(RecordSpecTreeXMLParser parser) throws XMLStreamException {
		parseXmlAttributesFrom(parser);
		parser.addSpecById(this);
		parseXmlChildrenFrom(parser);
	}
	
	/**
	 * @see ElementSpec#parseXmlAttributesFrom(RecordSpecTreeXMLParser)
	 */
	@Override
	protected void parseXmlAttributesFrom(RecordSpecTreeXMLParser parser) throws XMLStreamException {
		super.parseXmlAttributesFrom(parser);
		
		String fieldType = parser.getAttributeValue("fieldType");
		if (fieldType.equalsIgnoreCase("String")) {
			field = parser.getFactory().getStringField(
					parser.getAttributeIntValue("length"),
					parser.getAttributeBooleanValue("trim", true),
					parser.getAttributeBooleanValue("padLeft", false));
			
		} else if (fieldType.equalsIgnoreCase("Binary")) {
			precision = parser.getAttributeIntValue("precision");
			isComp5 = parser.getAttributeBooleanValue("comp5",false);
			field = parser.getFactory().getBinaryField(
					precision,
					parser.getAttributeIntValue("scale"), 
					parser.getAttributeBooleanValue("signed", true),
					isComp5);
			
		} else if (fieldType.equalsIgnoreCase("PackedDecimal")) {
			precision = parser.getAttributeIntValue("precision");
			field = parser.getFactory().getPackedDecimalField(
					precision,
					parser.getAttributeIntValue("scale"), 
					parser.getAttributeBooleanValue("signed", true));

		} else if (fieldType.equalsIgnoreCase("ExternalDecimal")) {
			precision = parser.getAttributeIntValue("precision");
			boolean signed = parser.getAttributeBooleanValue("signed", true);
			field = parser.getFactory().getExternalDecimalField(
					precision,
					parser.getAttributeIntValue("scale"), 
					signed,
					signed && parser.getAttributeBooleanValue("signTrailing", false),
					signed && parser.getAttributeBooleanValue("signExternal", true),
					parser.getAttributeBooleanValue("blankWhenZero", false));
			
		} else if (fieldType.equalsIgnoreCase("ExternalFloat")) {
			precision = parser.getAttributeIntValue("precision");
			field = parser.getFactory().getExternalFloatField(
					precision, 
					parser.getAttributeIntValue("scale"), 
					parser.getAttributeBooleanValue("showMantissaPlusSign", true),
					parser.getAttributeBooleanValue("showExponentPlusSign", true),
					parser.getAttributeBooleanValue("impliedDecimal", false));
			
		} else if (fieldType.equalsIgnoreCase("IbmFloat")) {
			field = parser.getFactory().getIbmFloatField();
			
		} else if (fieldType.equalsIgnoreCase("IbmDouble")) {
			field = parser.getFactory().getIbmDoubleField();
			
		} else if (fieldType.equalsIgnoreCase("ByteArray")) {
			field = parser.getFactory().getByteArrayField(
					parser.getAttributeIntValue("length"));
			
		}
	}		
}
