/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007, 2012. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.cobol;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Iterator;
import java.util.List;
import java.util.StringTokenizer;

import com.ibm.jzos.fields.*;
import com.ibm.jzos.recordgen.LibraryUtil;
import com.ibm.jzos.recordgen.LineFormatter;
import com.ibm.jzos.recordgen.asm.RecordSpecXMLParser;

/**
 * Generates the source code for a Java class that maps a byte array matching
 * the description given by a COBOL copy book.  The input to this
 * process is an ADATA file created by IBM Enterprise COBOL Compiler, or an XML file
 * that was created by {@link RecordXMLGenerator} from an ADATA file.
 * <p>
 * Refer to the "JZOS COBOL Record Generator User's Guide" and the sample JCL member 
 * "COBGEN" for example JCL for more information. 
 * <p>
 * @see #main(String[]) main() method for information on running this class
 * @since 2.4.4 support for parsing XML files (xmlfile= argument)
 * @see RecordXMLGenerator
 */
public class RecordClassGenerator {

	private static final String _TYPES_PACKAGE_NAME = "com.ibm.jzos.fields";
	
	/**
	 * Main method / driver for generating Java Classes from COBOL ADATA or XML files.
	 * Java source code is written to System.out.
	 * <p>
	 * Arguments consist of the following optional "name=value" pairs:  
	 * <ul>
	 * <li>adatafile=//MVS.DSN  (default: //DD:SYSADATA)</li>
	 * <li>xmlfile=//MVS.DSN    (default: none; exclusive with adatafile)</li> 
	 * <li>outputDir=dirname    (default: writes file to System.out (stdout))</li>
	 * <li>symbol=dataName      (default: first group found</li>
	 * <li>package=packageName  (default: "mypackage")</li>
	 * <li>class=className      (default: top-level group name</li>
	 * <li>genSetters=true|false (default: true)</li>
	 * <li>genCache=true|false  (default: true)</li>
	 * <li>bufoffset=true|false (default: false)</li>
	 * <li>stringTrim=true|false (default: false)</li>
	 * <li>stringEncoding=encoding (default: IBM-1047)</li>
	 * <li>preInitialize=true|false (default: false)</li>
	 * <li>genAccessorJavadoc=true|false (default: false)</li>
	 * </ul>
	 * <p> 
	 * If <code>outputDir</code> is given, the output file is created in the package-qualified
	 * subdirectory with a file name "Classname.java".  For example, if 
	 * <code>outputDir=/u/myid</code> and <code>package=com.myco.cobrecs</code> and
	 * <code>class=MyRecord</code> or implicitly determined from the copybook record 
	 * <code>MY-RECORD</code>, then the output file created will be: 
	 * <code>/u/myid/com/myco/cobrecs/MyRecord.java</code>
	 * <p>
	 * If the given <code>adatafile</code> is a pathname not starting with "//", then it is assumed
	 * to be the name of a binary file in the USS (or workstation) filesystem containing the 
	 * binary SYSADATA file.  In order to process SYSADATA files on the workstation, 
	 * they must be downloaded with IBM "RDWs" included.  This can be accomplished using FTP 
	 * along with the "bin" and "quote site rdw" commands".
	 * <p>
	 * If the given <code>adatafile</code> is a PDS name without a member or a directory name,
	 * then all members of the PDS (or files in the directory) are processed.  For this to work
	 * properly, the <code>outputDir</code> option must be specified but the 
	 * <code>class</code> and <code>symbol</code> options must not be specified so that the copybook top-level
	 * record name is used to generate unique classnames for each class.   This feature allows
	 * for one invocation of the RecordClassGenerator to generate Java classes for an
	 * entire PDS library of SYSADATA files.
	 * <p>
	 * If the name of an <code>xmlfile</code> is given, it is either the <code>//dataset.name</code>
	 * or <code>/path/name</code> of the file containing XML that was generated by or compatible
	 * with the output of {@link RecordXMLGenerator}.   If the <code>xmlfile</code> keyword is 
	 * specified, then the <code>adatafile</code> and <code>symbol</code> keywords may not be specified.
	 * <p>
	 * If  <code>preInitialize</code> is set to true, then code is generated 
	 * in the <code>setInitalValues()</code> method to initialize fields with a fixed location and length
	 * that are not arrays to blanks or zero.
	 * <p>
	 * @param args key=value strings as described above.
	 * @throws IOException
	 */
	public static void main(String[] args) throws IOException {
		
		String adatafile = "//DD:SYSADATA";
		boolean adatafileGiven = false;
		String xmlfile = null;
		String symbol = null;
		String packageName = "mypackage";
		String className = null;
		String outputDir = null;  
		boolean useBufOffset = false;
		boolean genSetters = true;
		boolean genCache = true;
		boolean stringTrim = false;
		String  stringEncoding = null;
		boolean genAccessorJavadoc = false;
		boolean preInitialize = false;
		
		for (int i = 0; i<args.length; i++) {
			StringTokenizer strtok = new StringTokenizer(args[i],"=");
			if (strtok.countTokens() != 2) {
				throw new IllegalArgumentException("Argument '" + args[i] + "' not in the form: name=value");
			}
			String name = strtok.nextToken();
			String value = strtok.nextToken();
			if ("adatafile".equalsIgnoreCase(name)) {
				adatafile = value;
				adatafileGiven = true;
			} else if ("xmlfile".equalsIgnoreCase(name)) {
				xmlfile = value;
			} else if ("outputDir".equalsIgnoreCase(name)) {
				outputDir = value;
			} else if ("symbol".equalsIgnoreCase(name)) {
				symbol = value;
			} else if ("package".equalsIgnoreCase(name)) {
				packageName = value;
			} else if ("class".equalsIgnoreCase(name)) {
				className = value;
			} else if ("genSetters".equalsIgnoreCase(name)) {
				genSetters = "true".equalsIgnoreCase(value);
			} else if ("gencache".equalsIgnoreCase(name)) {
				genCache = "true".equalsIgnoreCase(value);
			} else if ("bufOffset".equalsIgnoreCase(name)) {
				useBufOffset = "true".equalsIgnoreCase(value);
			} else if ("stringTrim".equalsIgnoreCase(name)) {
				stringTrim = "true".equalsIgnoreCase(value);
			} else if ("stringEncoding".equalsIgnoreCase(name)) {
				stringEncoding = value;
			} else if ("preInitialize".equalsIgnoreCase(name)) {
				preInitialize = "true".equalsIgnoreCase(value);
			} else if ("genAccessorJavadoc".equalsIgnoreCase(name)) {
				genAccessorJavadoc = "true".equalsIgnoreCase(value);
			} else {
				throw new IllegalArgumentException("Argument '" + name + "' not recognized.");
			}
		}

		if (xmlfile != null && (adatafileGiven || symbol != null)) {
			throw new IllegalArgumentException("xmlfile keyword may not be specified with either the adatafile or symbol keywords");
		}
		// Create a generator and configure it with the control arguments
		RecordClassGenerator rcg = new RecordClassGenerator();
		rcg.setClassName(className);
		rcg.setPackageName(packageName);
		rcg.setUseBufOffset(useBufOffset);
		rcg.setGenSetters(genSetters);
		rcg.setGenCache(genCache);
		rcg.setStringTrimDefault(stringTrim);
		rcg.setStringEncoding(stringEncoding);
		rcg.setPreInitialize(preInitialize);
		rcg.setGenAccessorJavadoc(genAccessorJavadoc);
		
		if (xmlfile != null) {
			processXMLFile(xmlfile, outputDir, rcg);
		} else {
			processLibrary(adatafile, outputDir, rcg, symbol);
		}
	}
	/**
	 * Return a File used to write to a generated output java source
	 * file, given the outputDir and class and package names from the given RecordClassGenerator.
	 * If the directory name is null, then return a BufferedWriter to System.out.
	 * @param rcg RecordClassGenerator used to determine the className and packageName; 
	 * @param outputDir the name of the output directory
	 */
	public static File openOutputFile(RecordClassGenerator rcg, String outputDir) throws IOException {
		
		if (outputDir == null) {
			return null; 
		}
		if (rcg.getClassName() == null) {
			throw new IllegalArgumentException("Given RecordClassGenerator does not have a className or specTree");
		}
		outputDir = LibraryUtil.resolveLeadingTilde(outputDir);
		File dir = new File(outputDir);
		if (!dir.exists()) {
			throw new IllegalArgumentException("Output directory \"" + outputDir + "\" does not exist");
		}
		if (!dir.isDirectory()) {
			throw new IllegalArgumentException("Output directory \"" + outputDir + "\" is not a directory");
		}
		String packageName = rcg.getPackageName().replace('.', '/');
		dir = new File(dir, packageName);
		if (!dir.exists()) {
			if (!dir.mkdirs()) {
				throw new IOException("Unable to make directory: " + dir.getCanonicalPath());
			}
		}
		return  new File(dir, rcg.getClassName() + ".java");
		
	}
	/**
	 * Return the unqualified (simple) name of a Class
	 * @param c the Class
	 * @return String
	 */
	private static String getUnqualifiedClassname(Class c) {
		String fqn = c.getName();
		int i = fqn.lastIndexOf('.');
		if (i < 0) return fqn;
		return fqn.substring(i+1);
	}  

	/**
	 * Process a single SYSADATA file
	 * @param adatafile the file name.  If starts with "//", then it is assumed to be 
	 * a z/OS dataset
	 * @param outputDir the output directory name where the output class is written, 
	 * under its package-qualified subdirectories
	 * @param rcg the RecordClassGenerator which has the class and package name
	 * @param symbol the name of the top-level record (group) data item to process.  
	 * If null, then process the first encountered.
	 */
	private static void processFile(String adatafile, String outputDir, RecordClassGenerator rcg, String symbol) {
		
		BufferedWriter wtr = null;
		try {
			// First, a RecordSpecBuilder converts the selected symbol of the ADATA
			// file into a RecordSpec model.
			adatafile = LibraryUtil.resolveLeadingTilde(adatafile);
			RecordSpecTreeBuilder specBuilder = new RecordSpecTreeBuilder();
			specBuilder.setSelectedSymbolName(symbol);
			StorageSpec specTree = specBuilder.process(adatafile);
			rcg.setSpecTree(specTree);
		
			// Generate Java source to System.out
			File f = openOutputFile(rcg, outputDir);
			String outName;
			if (f != null) {
				wtr = new BufferedWriter(new FileWriter(f));
				outName = f.getCanonicalPath();
			} else {
				wtr = new BufferedWriter(new OutputStreamWriter(System.out));
				outName = "System.out (stdout)";
			}
			System.err.println("Generating " + adatafile + " to: " + outName);
			rcg.generateJavaSourceOn(wtr);
			
		} catch (Throwable t) {
			System.err.println("Error while processing file: " + adatafile + "\n\t" + t);
			t.printStackTrace();
		} finally {
			try { // to close the writer
				if (wtr != null) wtr.close(); 
			} catch(Throwable ignore) {}
		}
	}

	/**
	 * Process a single XML file
	 * @param xmlfile the file name.  If starts with "//", then it is assumed to be 
	 * a z/OS dataset
	 * @param outputDir the output directory name where the output class is written, 
	 * under its package-qualified subdirectories
	 * @param rcg the RecordClassGenerator which has the class and package name
	 */
	private static void processXMLFile(String xmlfile, String outputDir, RecordClassGenerator rcg) {
		
		BufferedWriter wtr = null;
		try {
			// First, a RecordSpecBuilder converts the selected symbol of the ADATA
			// file into a RecordSpec model.
			xmlfile = LibraryUtil.resolveLeadingTilde(xmlfile);
			RecordSpecTreeXMLParser specParser = new RecordSpecTreeXMLParser();
			StorageSpec specTree = specParser.process(xmlfile);
			rcg.setSpecTree(specTree);
		
			// Generate Java source to System.out
			File f = openOutputFile(rcg, outputDir);
			String outName;
			if (f != null) {
				wtr = new BufferedWriter(new FileWriter(f));
				outName = f.getCanonicalPath();
			} else {
				wtr = new BufferedWriter(new OutputStreamWriter(System.out));
				outName = "System.out (stdout)";
			}
			System.err.println("Generating " + xmlfile + " to: " + outName);
			rcg.generateJavaSourceOn(wtr);
			
		} catch (Throwable t) {
			System.err.println("Error while processing file: " + xmlfile + "\n\t" + t);
			t.printStackTrace();
		} finally {
			try { // to close the writer
				if (wtr != null) wtr.close(); 
			} catch(Throwable ignore) {}
		}
	}
	
	/**
	 * Process all of the members of a given adatafile if it is a PDS, or all of the
	 * files if it is a directory, or a single file otherwise.
	 */
	private static void processLibrary(String adatafile, String outputDir, RecordClassGenerator rcg, String symbol) throws IOException {

		// Get the the fully qualified member file names in the library, 
		// which would be a single name if the adatafile was not a library or directory
		List files = LibraryUtil.getLibraryMembers(adatafile);
		
		if (files.size() == 0) {
			throw new IllegalArgumentException("The given adatafile PDS or directory was empty");
		}
		if (files.size() > 1) {
			if (rcg.getClassName() != null) {
				System.err.println("class=className argument cannot be specified if a adatafile PDS or directory is processed and is ignored");
				rcg.setClassName(null);
			}
			if (symbol != null) {
				System.err.println("symbol=name argument cannot be specified if a adatafile PDS or directory is processed and is ignored");
				symbol = null;
			}
			if (outputDir == null) {
				throw new IllegalArgumentException("The outputDir= argument must be supplied when processing a PDS or directory");
			}
		}
		for (Iterator i=files.iterator(); i.hasNext(); ) {
			String filename = (String)i.next();
			processFile(filename, outputDir, rcg, symbol);
		}	
	}
	
	/** The tree of ElementSpecs, rooted in a StorageSpec, built by SpecTreeBuilder */
	private StorageSpec specTree;	

	/** The name of the output class.  If null defaults to the top-level record name.
	 * @see getClassName()
	 */
	private String className = null;

	// Note: the following state ivs should be reinitialized when the specTree is reset.
	private LineFormatter formatter;
	private CobolDatatypeFactory factory = new CobolDatatypeFactory();
	private ArrayList fieldSpecsWithInitialValues = new ArrayList();
	private ArrayList innerClassSpecs = new ArrayList();  // of StorageSpecs that will gen inner classes
	private boolean innerClass = false;
	
	// configuration properties
	/** Whether some DEBUG comments are generated in the output class */
	private boolean genDebugComments = false;
	
	/** Whether static fields are generated protected (true) or public(false) */
	private boolean genProtectedFields = true;
	
	/** Whether sourcelines are used as javadoc for getters and setters */
	private boolean genAccessorJavadoc = false;
	
	/** Whether field setter methods are generated */
	private boolean genSetters = true;
	
	/** Whether instances variables are used as field cache values */
	private boolean genCache = true;

	/** The default setting for {@link DatatypeFactory#setStringTrimDefault(boolean)} */
	private boolean stringTrimDefault = false;
	
	/** If true, code is generated to initialize fields of fixed location and length to blanks/zeros */
	private boolean preInitialize = false;
	

	/** The default String encoding for the DatatypeFactory.
	 * If not set, defaults to IBM-1047
	 */
	private String stringEncoding = null;
	
	/** The package name of the output class */
	private String packageName = "mypackage";
	
	/** Whether a class that uses bufOffset if generated */
	private boolean useBufOffset = false;
	
	/**
	 * Construct an empty instance.  
	 */
	public RecordClassGenerator() {
	}
	
	/**
	 * Construct an instance with a given StorageSpec (tree), which was created by
	 * a {@link RecordSpecTreeBuilder}
	 * @param aStorageSpec the StorageSpec tree
	 */
	public RecordClassGenerator(StorageSpec aStorageSpec) {
		setSpecTree(aStorageSpec);
	}
	
	/**
	 * Generate Java source to the given Writer.
	 * @param writer
	 * @throws IOException
	 */
	public void generateJavaSourceOn(Writer writer) throws IOException {

		validateSpecTree();		
		try {
			this.formatter = new LineFormatter(writer);
			findInnerClassSpecs(specTree);
			genFileHeader();
			genClassHeader();
			genStaticFieldLayout();
			genInstanceVariables();
			genConstructors();
			genBufferAccessors();
			genSetInitialValuesMethod();
			genGettersAndSetters();
			genInnerClasses();
			genClassTrailer();
		} finally {
			try {
				writer.flush();
			} catch(Throwable ignored) {};
		}
	}
	
	/**
	 * Answer the unqualified Java class name that is generated.
	 * If this is not given explicitly, then the top-level Cobol record (group) 
	 * data item is used to generate a Java class name.
	 * 
	 * @return String the unqualified ClassName
	 */
	public String getClassName() {
		if (className != null) {
			return className;
		} else if (specTree == null) {
			return null;
		} else {
			return getInstanceJavaName(specTree, true);
		}
	}
	
	/**
	 * Answer the Java package name for the generated Class.
	 * If not specified, this defaults to "mypackage".
	 * @return String the packagename
	 */
	public String getPackageName() {
		return packageName;
	}
	
	/**
	 * Answer the current root StorageSpec
	 * @return StorageSpec
	 */
	public StorageSpec getSpecTree() {
		return specTree;
	}
	
	/**
	 * Answer the default value for StringField trimming.
	 * @see DatatypeFactory#setStringTrimDefault(boolean)
	 */
	public boolean getStringTrimDefault() {
		return stringTrimDefault;
	}
	
	/**
	 * Set the default trim setting for StringFields
	 * @see #getStringTrimDefault()
	 */
	public void setStringTrimDefault(boolean value) {
		this.stringTrimDefault = value;
	}	

	/**
	 * Answer the default String encoding for the DatatypeFactory.
	 * If null, IBM-1047 is used.
	 */
	public String getStringEncoding() {
		return stringEncoding;
	}
	
	/**
	 * Set the default encoding for StringFields
	 * @see #getStringEncoding()
	 */
	public void setStringEncoding(String value) {
		this.stringEncoding = value;
	}	

	/**
	 * Answer whether code will be generated to pre-initialize non-Array fields
	 * with fixed location and length to blanks
	 */
	public boolean isPreInitialize() {
		return preInitialize;
	}
	
	/**
	 * Set whether code will be generated to pre initialize non-Array fields
	 * with fixed location and length to blanks
	 * @see #isPreInitStrings()
	 */
	public void setPreInitialize(boolean preInitialize) {
		this.preInitialize = preInitialize;
	}
	
	/**
	 * Answer whether special //DEBUG source comments are generated
	 * @return boolean
	 */
	public boolean isGenDebugComments() {
		return genDebugComments;
	}

	/**
	 * Answer whether Field declarations should be protected (default) or public.
	 * @return boolean
	 */
	public boolean isGenProtectedFields() {
		return genProtectedFields;
	}	

	/** 
	 * Return whether sourcelines are used as javadoc for getters and setters. 
	 */
	public boolean isGenAccessorJavadoc() {
		return genAccessorJavadoc;
	}
	
	/**
	 * Answer whether Java "set" methods should be generated for generated fields.
	 * @return boolean
	 */
	public boolean isGenSetters() {
		return genSetters;
	}

	/**
	 * Answer whether generated Java source code will include buffer offsets
	 * into the underlying byte array.
	 * @return boolean
	 */
	public boolean isUseBufOffset() {
		return useBufOffset;
	}

	/**
	 * @see #getClass()
	 */
	public void setClassName(String className) {
		this.className = className;
	}

	/**
	 * @see #isGenDebugComments()
	 */
	public void setGenDebugComments(boolean genDebugComments) {
		this.genDebugComments = genDebugComments;
	}

	/**
	 * @see #isGenProtectedFields()
	 */
	public void setGenProtectedFields(boolean genProtectedFields) {
		this.genProtectedFields = genProtectedFields;
	}

	/**
	 * @see #isGenAccessorJavadoc()
	 */
	public void setGenAccessorJavadoc(boolean genAccessorJavadoc) {
		this.genAccessorJavadoc = genAccessorJavadoc;
	}

	/**
	 * @see #isGenSetters()
	 */
	public void setGenSetters(boolean genSetters) {
		this.genSetters = genSetters;
	}
	
	/**
	 * Answer true if instances variables are used to cache field values
	 */
	public boolean isGenCache() {
		return this.genCache;
	}

	/**
	 * @see #isGenCache()
	 */
	public void setGenCache(boolean value) {
		this.genCache = value;
		if (specTree != null) {
			specTree.setGenerateCachedAccessors(value);
		}
	}
	
	/**
	 * @see #getPackageName()
	 */
	public void setPackageName(String packageName) {
		this.packageName = packageName;
	}


	/**
	 * Set the StorageSpec tree to process
	 */
	public void setSpecTree(StorageSpec specTree) {
		this.specTree = specTree;
		validateSpecTree();
		
		// if not caching, disable it for the entire specTree
		specTree.setGenerateCachedAccessors(isGenCache());

		// reset processing state
		formatter = null;
		innerClass = false;
		fieldSpecsWithInitialValues = new ArrayList();
		factory = new CobolDatatypeFactory();	
		innerClassSpecs = new ArrayList(); 
	}

	/**
	 * @see #isUseBufOffset()
	 */
	public void setUseBufOffset(boolean useBufOffset) {
		this.useBufOffset = useBufOffset;
	}
	

	/**
	 * Builds the a list of innerClassSpecs which are GroupSpecs with OCCURS clauses.
	 * @param storageSpec the subtree to search
	 * @throws IOException
	 */
	protected void findInnerClassSpecs(StorageSpec storageSpec) throws IOException {

		if (storageSpec.isSkipCodeGeneration()) return;
		if (!(storageSpec instanceof GroupSpec)) return;
		for (Iterator<ElementSpec> iter = storageSpec.getChildren().iterator(); iter.hasNext();) {
			ElementSpec spec = iter.next();
			if (spec instanceof GroupSpec) {
				GroupSpec groupSpec = (GroupSpec)spec;
				if (groupSpec.isArray()) {
					innerClassSpecs.add(groupSpec);
				}
				findInnerClassSpecs(groupSpec);
			}
		}		
	}

	/**
	 * Generate the method <code>getByteBuffer()</code> and, if
	 * {@link #isUseBufOffset()} is enabled, the <code>getByteBufferOffset()</code> method.
	 * @throws IOException
	 */
	protected void genBufferAccessors() throws IOException {		

		formatter.println();
		formatter.println("public byte[] getByteBuffer() {");
		formatter.println("return _byteBuffer;");
		formatter.println("}");

		if (useBufOffset) {
			formatter.println();
			formatter.println("public int getByteBufferOffset() {");
			formatter.println("return _byteBufferOffset;");
			formatter.println("}");
		}
	}

	/**
	 * Generates the <code>cacheOdoSizes()</code> method, which pre-initializes
	 * and OCCURS DEPENDING ON objects (sizes).
	 * @throws IOException
	 */
	protected void 	genCacheOdoSizesMethod() throws IOException {
		formatter.println();
		formatter.println("private void cacheOdoSizes() {");
		genOdoObjectInitializers(specTree);
		formatter.println("}");
	}
	
	/**
	 * Generate the header for a class
	 * @throws IOException
	 */
	protected void genClassHeader() throws IOException {
		
		formatter.println("public% class % {",
				innerClass ? " static" : "",
				getClassName()
				);
	}
	
	/**
	 * Generate the trailer for a class
	 * @throws IOException
	 */
	protected void genClassTrailer() throws IOException {

		formatter.println();
		formatter.println("}");
		
	}
	
	/**
	 * For ConditionSpecs, we generate a "is<conditionName>" method. 
	 */
	protected void genConditionMethod(ConditionSpec conditionSpec) throws IOException {
		
		if (conditionSpec.getValueClause() == null) {
			// one that we didn't support
			return;
		}
		FieldSpec fieldSpec = (FieldSpec)conditionSpec.getParent();
		Field field = fieldSpec.getField();
		String fieldJavaName = getInstanceJavaName(fieldSpec, true);
		String conditionConstName = getStaticJavaName(conditionSpec);
		String conditionJavaName = getInstanceJavaName(conditionSpec, true);
		
		String equals;
		String closeEquals;
		if (field instanceof IntAccessor
				|| field instanceof LongAccessor
				|| field instanceof FloatAccessor
				|| field instanceof DoubleAccessor ) {
			equals = " == ";
			closeEquals = "";
		} else {
			equals = ".equals(";
			closeEquals = ")";
		}
				
		
		formatter.println();
		formatter.println("public boolean is%() {",
				conditionJavaName);
		formatter.println("return get%()%%%;",
					fieldJavaName,
					equals,
					conditionConstName,
					closeEquals);
		formatter.println("}");
	}

	/** 
	 * Generate the constructors for a class, which vary depending on a number of factors:
	 * <ul>
	 * <li>Whether {@link #isUseBufOffset()} is enabled</li>
	 * <li>Whether the record contains ODOs</li>
	 * <li>Whether there are items with (initial) VALUEs clauses</li>
	 * </ul>
	 * @throws IOException
	 */
	protected void genConstructors() throws IOException {
	
		formatter.println();
		formatter.println();
		formatter.println("public % (byte[] buffer%%) {",
				getClassName(),
				useBufOffset ? ", int bufOffset" : "",
				specTree.containsOdo() ? ", boolean bufferContainsData" : "");

		formatter.println("this._byteBuffer = buffer;");
		if (useBufOffset) {
			formatter.println("this._byteBufferOffset = bufOffset;");
		}
		if (specTree.containsOdo()) {
			formatter.println("if (bufferContainsData) {");
			formatter.println("cacheOdoSizes();");
			formatter.println("}");
		}
		formatter.println("}");
		
		if (specTree.containsOdo()) {
			genCacheOdoSizesMethod();
			return; // don't gen default ctor
		}

		// the default ctor is generated only if we are fixed size (don't contain an odo)
		formatter.println();
		formatter.println("public % () {",
				getClassName());

		formatter.println("this._byteBuffer = new byte[%_len];",
							getStaticJavaName(specTree));

		if (fieldSpecsWithInitialValues.size() > 0) {
			formatter.println("setInitialValues();");
		}
		formatter.println("}");
	}

	/**
	 * Generate the file header for the class
	 * @throws IOException
	 */
	protected void genFileHeader() throws IOException {
		String pkg = getPackageName();
		if (pkg != null && pkg.trim().length() > 0) {
			formatter.println("package %;", getPackageName());
		}
		formatter.println("import %.*;", _TYPES_PACKAGE_NAME);
		if (((GroupSpec)specTree).containsBigDecimalsOrIntegers()) {
			formatter.println("import java.math.*;");
		}
		formatter.println();
		formatter.println("// Generated by % on %",
				getClass().getName(),
				Calendar.getInstance().getTime());
		formatter.println();
		
	}

	/**
	 * Generate the getter and setter methods for a FieldSpec
	 * @param fieldSpec
	 * @throws IOException
	 */
	protected void genGetterAndSetter(FieldSpec fieldSpec) throws IOException {

		Field field = fieldSpec.getField();
		String symbol = fieldSpec.getSymbol();
		if (symbol == null) return; 
		
		String javaName = getInstanceJavaName(fieldSpec, true);
		String argName = getInstanceJavaName(fieldSpec, false);
		String ivName = argName;
		String fieldName = getStaticJavaName(fieldSpec);
		String fieldGetter = "";
		String fieldSetter = "";
		String typeName;
		String typeGetter = "";
		String ivConstructor = "";
				
		// Generate array-indexes accessors
		if (fieldSpec.isArray()) {
			ivName = ivName + "[_arrayIndex]";
		}

		// write arguments to factory "constructor", by type
		if (field instanceof StringField ) {
			typeName = "String";
			fieldGetter = ".getString";
			fieldSetter = ".putString";
			
		} else if (field instanceof IntAccessor ) {
			typeName = "int"; 
			typeGetter = ".intValue()";
			fieldGetter = ".getInt";
			fieldSetter = ".putInt";
			ivConstructor = "new Integer(";
			
		} else if (field instanceof LongAccessor ) {
			typeName = "long";
			typeGetter = ".longValue()";
			fieldGetter = ".getLong";
			fieldSetter = ".putLong";
			ivConstructor = "new Long(";
			
		} else if (field instanceof FloatAccessor ) {
			typeName = "float";
			typeGetter = ".floatValue()";
			fieldGetter = ".getFloat";
			fieldSetter = ".putFloat";
			ivConstructor = "new Float(";
			
		} else if (field instanceof DoubleAccessor ) {
			typeName = "double";
			typeGetter = ".doubleValue()";
			fieldGetter = ".getDouble";
			fieldSetter = ".putDouble";
			ivConstructor = "new Double(";
			
		} else if (field instanceof ByteArrayField ) {
			typeName = "byte[]";
			fieldGetter = ".getByteArray";
			fieldSetter = ".putByteArray";
			
		} else if (field instanceof BigDecimalAccessor ) {
			typeName = "BigDecimal";
			fieldGetter = ".getBigDecimal";
			fieldSetter = ".putBigDecimal";
			
		} else if (field instanceof BigIntegerAccessor ) {
			typeName = "BigInteger";
			fieldGetter = ".getBigInteger";
			fieldSetter = ".putBigInteger";
			
		} else {  
			return; // unsupported
		}

		// Prebuild the offset arguments if needed
		StringBuffer offsets = new StringBuffer();
		if (fieldSpec.isVarLoc()) {  // if field follows and ODO
			offsets.append("_variableOffset");
		}
		if (fieldSpec.isArray()) {
			if (offsets.length() > 0) offsets.append(" + ");
			offsets.append("_arrayOffset");
		}
		if (useBufOffset) {
			if (offsets.length() > 0) offsets.append(" + ");
			offsets.append("_byteBufferOffset");
		}
		
		// Gen the getter...
		
		formatter.println();
		if (genAccessorJavadoc) {
			genSourceLineComments(fieldSpec);
		}
		//  public <typeName> get<javaName>(<int _arrayIndex>) {
		formatter.println("public % get%(%) {",
				typeName,
				javaName,
				fieldSpec.isArray() ? "int _arrayIndex" : "");
		
		if (fieldSpec.isGenerateCachedAccessors()) {
			formatter.println("if (% == null) {", ivName);
		}
		
		if (fieldSpec.containsOdoObject()) {
			//      throw new IllegalStateException("ODO object not initialized");
			formatter.println("throw new IllegalStateException(\"ODO object not initialized\");");
			
		} else {
			// For VariableLength StringFields, dynamically create a StringField instance
			if (fieldSpec.isVarLenString()) {
				//   StringField _field = new StringField(<fieldName>.getOffset(), get<odoObject>(), false);
				OccursDependingOn odo = (OccursDependingOn)fieldSpec.getOccurs();
				StorageSpec odoObject = odo.getOdoObject();
				formatter.println("StringField _field = new StringField(%.getOffset(), get%(), false);",
						fieldName,
						getInstanceJavaName(odoObject, true));
				fieldName = "_field";  // use this as the name of the field converter
			}
			
			// (for varloc fields)
			//      int _variableOffset = get<prevOdoJavaName>_EndOffset();
			if (fieldSpec.isVarLoc()) {
				VariableOffset vOffset = (VariableOffset)fieldSpec.getOffset();
				StorageSpec previousOdo = vOffset.getOdo();
				formatter.println("int _variableOffset = get%_EndOffset();",
						getInstanceJavaName(previousOdo, true));
			}
			
			// (for arrays)
			//      int _arrayOffset = _arrayIndex * <fieldName>.getByteLength();
			if (fieldSpec.isArray()) {
				formatter.println("int _arrayOffset = _arrayIndex * %.getByteLength();",
						fieldName);
			}
	
			//      <ivName = <ivConstructor><fieldName><fieldGetter>(_byteBuffer <,varOffset <+ _byteBufferOffset <+ _arrayOffset>>>)<)>;
			formatter.println("%%%(_byteBuffer%)%;",
					fieldSpec.isGenerateCachedAccessors()
						? ivName + " = " + ivConstructor
						: "return ",
					fieldName,
					fieldGetter,
					offsets.length() > 0 ? ", " + offsets : "",
					(ivConstructor.length() > 0 && fieldSpec.isGenerateCachedAccessors()) ? ")" : "");
		}

		if (fieldSpec.isGenerateCachedAccessors()) {
			//    }
			formatter.println("}");
			
			//    return <ivName><typeGetter>;
			formatter.println("return %%;", ivName, typeGetter);
		}

		//    }
		formatter.println("}");
		
		//>>>> Gen the setter  <<<<<
		if (!genSetters) return;
		fieldName = getStaticJavaName(fieldSpec); 
		
		formatter.println();
		if (genAccessorJavadoc) {
			genSourceLineComments(fieldSpec);
		}
		//  public void set<javaName>(<int _arrayIndex,> <typeName> <argName>) {
		formatter.println("public void set%(%% %) {",
				javaName,
				fieldSpec.isArray() ? "int _arrayIndex, " : "",
				typeName,
				argName);
		
		if (fieldSpec.isGenerateCachedAccessors()) {
			//    if (<fieldName>.equals(this.<ivName>, <argName>)) {
			//      return;
			//    }
			formatter.println("if (%.equals(this.%, %)) {",
					fieldName,
					ivName,
					argName);
			formatter.println("return;");
			formatter.println("}");
		}
		
		// (for ODO objects)
		if (fieldSpec.containsOdoObject()) {
			//  if (this.<fieldName> != null) {
			//      throw new IllegalStateException("ODO object size may not be altered once set");
			//  }
			formatter.println("if (this.% != null) {", ivName);
			formatter.println("throw new IllegalStateException(\"ODO object size may not be altered once set\");");
			formatter.println("}");
		}
		
		// For variable length StringFields (ODOs)
		if (fieldSpec.isVarLenString()) {
			//   StringField _field = new StringField(<fieldName>.getOffset(), get<odoObject>(), false);
			OccursDependingOn odo = (OccursDependingOn)fieldSpec.getOccurs();
			StorageSpec odoObject = odo.getOdoObject();
			formatter.println("StringField _field = new StringField(%.getOffset(), get%(), false);",
					fieldName,
					getInstanceJavaName(odoObject, true));
			fieldName = "_field";
		}
		
		// (for varloc fields)
		//      int _variableOffset = get<prevOdoJavaName>_EndOffset();
		if (fieldSpec.isVarLoc()) {
			VariableOffset vOffset = (VariableOffset)fieldSpec.getOffset();
			StorageSpec previousOdo = vOffset.getOdo();
			formatter.println("int _variableOffset = get%_EndOffset();",
					getInstanceJavaName(previousOdo, true));
		}
		
		// (for arrays)
		//    int _arrayOffset = _arrayIndex * <fieldName>.getByteLength();
		if (fieldSpec.isArray()) {
			formatter.println("int _arrayOffset = _arrayIndex * %.getByteLength();",
					fieldName);
		}		
		
		//    <fieldName><fieldSetter>(<argName>, _byteBuffer<,varOffset <+ _byteBufferOffset <+ _arrayOffset>>>);
		formatter.println("%%(%, _byteBuffer%);",
				fieldName,
				fieldSetter,
				argName,
				offsets.length() > 0 ? ", " + offsets : "");
		
		if (fieldSpec.isGenerateCachedAccessors()) {
			//    this.<ivName> = <ivConstructor><argName><)>;
			formatter.println("this.% = %%%;",
					ivName, 
					ivConstructor,
					argName,
					ivConstructor.length() > 0 ? ")" : "");
		}		
		formatter.println("}");

		// If this is an ODO field, generate the code needed to determine the
		// end offset at runtime.
		if (fieldSpec.isOdo()) {
			genOdoEndOffsetMethod(fieldSpec);
		}
	
	}
	
	/**
	 * Generate getter and setter methods for a GroupSpec
	 * @param groupSpec the GroupSpec tree
	 * @return true if the children of the GroupSpec should be traversed
	 * @throws IOException
	 */
	protected boolean genGetterAndSetter(GroupSpec groupSpec) throws IOException {

		// If the group doesn't have an OCCURS or ODO clause, children need to
		// be visited.  For groups that generate StringFields, generate the accessors
		if (!groupSpec.isArray()) {
			if (groupSpec.isGenerateStringField()) {
				FieldSpec stringFieldSpec = new FieldSpec();
				stringFieldSpec.setSymbol(groupSpec.getSymbol());
				stringFieldSpec.setOffset(groupSpec.getOffset());
				stringFieldSpec.setField(new StringField(0,groupSpec.getSize()));
				stringFieldSpec.setGenerateCachedAccessors(groupSpec.isGenerateCachedAccessors());
				genGetterAndSetter(stringFieldSpec);
			}
			return true; // -> continue traversing children
			
		// If the group has an OCCURS or ODO clause then generate a getter
		// for indexed inner class instances.  Also, stop traversing at this point.
		} else {
			
			genInnerClassIndexedGetter(groupSpec);
			
			// If this is an ODO group, generate the code needed to determine the
			// end offset at runtime.
			if (groupSpec.isOdo()) {
				genOdoEndOffsetMethod(groupSpec);
			}
			return false;  // -> don't traverse children
		}

	}

	/**
	 * Generate the getter and setter methods for a StorageSpec subtree.
	 * @param storageSpec the subtree to visit
	 * @throws IOException
	 */
	protected void genGetterAndSetter(StorageSpec storageSpec) throws IOException {
		
		if (storageSpec.isSkipCodeGeneration()) return;
		boolean traverseChildren = true;
		if (storageSpec instanceof GroupSpec) {
			traverseChildren = genGetterAndSetter((GroupSpec)storageSpec);
		} else {
			genGetterAndSetter((FieldSpec)storageSpec);			
		}
		
		if (traverseChildren) {
			for (Iterator<ElementSpec> iter = storageSpec.getChildren().iterator(); iter.hasNext();) {
				ElementSpec spec = iter.next();
				
				if (spec instanceof StorageSpec) {
					genGetterAndSetter((StorageSpec)spec);
				} else if (spec instanceof ConditionSpec) {
					genConditionMethod((ConditionSpec)spec);
				} else {
					//RenameSpecs are not currently supported
				}
			}
		}
	}

	/** 
	 * Generate all of the getters and setters
	 * @throws IOException
	 */
	protected void genGettersAndSetters() throws IOException {
		formatter.println();
		
		genGetterAndSetter(specTree);
	}
	
	/**
	 * Generate static inner classes for any GroupSpecs with OCCURS clauses
	 * @throws IOException
	 */
	protected void genInnerClasses() throws IOException {
		
		boolean saveBufOffset = useBufOffset;
		useBufOffset = true;  // inner classes must use buffer offsets
		innerClass = true;
		String saveClassName = this.className;
		
		for (Iterator i=innerClassSpecs.iterator(); i.hasNext(); ) {
			GroupSpec groupSpec = (GroupSpec)i.next();
			this.className = getInstanceJavaName(groupSpec, true);
			this.specTree = groupSpec;
			groupSpec.setOccurs(null);  // the inner class shouldn't know that it is an array
			groupSpec.setOffset(new Offset(0));  // always relative to 0 in inner class gen
			groupSpec.setVarLoc(false);
			factory.setOffset(0);
			fieldSpecsWithInitialValues = new ArrayList(); // reset for each inner class
			genSourceLineComments(groupSpec);
			genClassHeader();
			genStaticFieldLayout();
			genInstanceVariables();
			genConstructors();
			genBufferAccessors();
			genSetInitialValuesMethod();
			genGettersAndSetters();
			genClassTrailer();
		}
		
		innerClass = false;
		useBufOffset = saveBufOffset;
		className = saveClassName;
	}	

	/**
	 * Generate an array-indexed getter for a specific GroupSpec (one with an OCCURS clause)
	 * @param groupSpec
	 * @throws IOException
	 */
	protected void genInnerClassIndexedGetter(GroupSpec groupSpec) throws IOException {

		String innerClassName = getInstanceJavaName(groupSpec, true);
		String ivName = getInstanceJavaName(groupSpec, false) + "[_arrayIndex]";
		String fieldName = getStaticJavaName(groupSpec);
		
		//  public <innerClassName> get<javaName>(int _arrayIndex) {
		formatter.println("public % get%(int _arrayIndex) {",
				innerClassName,
				innerClassName);
		
		if (groupSpec.isGenerateCachedAccessors()) {
			//    if (<ivName> == null) {
			formatter.println("if (% == null) {", ivName);
		}
		
		// (for varloc fields)
		//      int _variableOffset = get<prevOdoJavaName>_EndOffset();
		if (groupSpec.isVarLoc()) {
			VariableOffset vOffset = (VariableOffset)groupSpec.getOffset();
			StorageSpec previousOdo = vOffset.getOdo();
			formatter.println("int _variableOffset = get%_EndOffset();",
					getInstanceJavaName(previousOdo, true));
		}
		
		//      int _arrayOffset = _arrayIndex * <fieldName>_len;
		formatter.println("int _arrayOffset = _arrayIndex * %_len;", fieldName);

		//      <ivName> = new <innerClassName>(_byteBuffer ,<fieldName>_len + _arrayOffset <+ varOffset <+ _byteBufferOffset>>);
		//    }	
		if (groupSpec.isGenerateCachedAccessors()) {
			formatter.print("% = ", ivName);
		} else {
			formatter.print("return ");
		}
		
		formatter.println("new %(_byteBuffer, %%_offset + _arrayOffset%%);",
				innerClassName,
				useBufOffset ? "_byteBufferOffset + " : "",
				fieldName,
				groupSpec.isVarLoc() ? " + _variableOffset" : "");
		
		if (groupSpec.isGenerateCachedAccessors()) {
			//    }
			//    return <ivName>;
			formatter.println("}");
			formatter.println("return %;", ivName);
		}

		//  }
		formatter.println("}");
	}
	
	/**
	 * Generate an instance variable declaration used for caching the value of a FieldSpec
	 * @param fieldSpec the FieldSpec
	 * @throws IOException
	 */
	protected void genInstanceVariable(FieldSpec fieldSpec) throws IOException {

		
		Field field = fieldSpec.getField();
		String javaName = getInstanceJavaName(fieldSpec, false);
		String fieldName = getStaticJavaName(fieldSpec);

		String ivClassName;

		// write arguments to factory "constructor", by type
		if (field instanceof StringField ) {
			ivClassName = "String";
			
		} else if (field instanceof IntAccessor ) {
			ivClassName = "Integer"; 
			
		} else if (field instanceof LongAccessor ) {
			ivClassName = "Long";
			
		} else if (field instanceof BigDecimalAccessor ) {
			ivClassName = "BigDecimal";
			
		} else if (field instanceof BigIntegerAccessor ) {
			ivClassName = "BigInteger";
			
		} else if (field instanceof FloatAccessor ) {
			ivClassName = "Float";
			
		} else if (field instanceof DoubleAccessor ) {
			ivClassName = "Double";
			
		} else if (field instanceof ByteArrayField ) {
			ivClassName = "byte[]";
			
		} else {  
			return; // unsupported
		}
		
		genInstanceVariableCode(javaName, fieldName, ivClassName, fieldSpec.isArray());
	}
	
	/**
	 * Generate the instance variable for a GroupSpec. 
	 * @param groupSpec
	 * @return true if the children of the GroupSpec should be traversed
	 * @throws IOException
	 */
	protected boolean genInstanceVariable(GroupSpec groupSpec) throws IOException {
	
		String javaName = getInstanceJavaName(groupSpec, false);
		String fieldName = getStaticJavaName(groupSpec);		
		String ivClassName;
		boolean isArray = false;
		
		// If the group is an array, then generate an iv.
		// to hold a cached array of inner class instances
		if (groupSpec.isArray()) {
			
			// get the name of the inner class
			ivClassName = getInstanceJavaName(groupSpec, true);
			isArray = true;
			
		// If the group isn't an array, then it might have a StringField overlay
		} else if (groupSpec.isGenerateStringField()) {
			ivClassName = "String";
			
		// Otherwise, we don't have an instance variable
		} else {
			return true;  // -> continue traversing children
		}
		
		genInstanceVariableCode(javaName, fieldName, ivClassName, isArray);
		
		// Answer true (traverse children) if we don't have an OCCURS/ODO clause
		return !groupSpec.isArray();
		
	}

	/**
	 * Generate code for a instance variable declaration
	 * @param javaName the java variable name
	 * @param fieldName for arrays, the static field name
	 * @param ivClassName for arrays, the type name of the array
	 * @param isArray true if the iv is generated as an array
	 * @throws IOException
	 */
	protected void genInstanceVariableCode(String javaName, String fieldName, 
										String ivClassName, boolean isArray) throws IOException {
		
		//  protected <ivClassName> <javaName>;   
		//     or, for arrays:
		//  protected <ivClassName>[] <javaName>= new <ivClassName>[<fieldName>_num];
		formatter.print("protected %% %",
				ivClassName,
				isArray ? "[]" : "",
				javaName);
		if (isArray) {
			formatter.print(" = new %[%_num]",
					ivClassName,
					fieldName);
		}
		formatter.println(";");
	}
	
	
	/**
	 * Generate all of the instance variables for a class.  These include the
	 * <code>_byteBuffer</code>, possibly the <code>_byteBufferOffset</code>
	 * and all of the instance variables used for field caching.
	 * @throws IOException
	 */
	protected void genInstanceVariables() throws IOException {

		formatter.println();
		formatter.println("protected byte[] _byteBuffer;");
		if (useBufOffset) {
			formatter.println("protected int _byteBufferOffset;");
		}
		if (isGenCache()) {
			formatter.println("// Instance variables used to cache field values");
		}
		// We continue since ODO objects *must* be cached
		genInstanceVariables(specTree);
	}
	
	/**
	 * Generate the instance variables for a StorageSpec subtree
	 * @param storageSpec
	 * @throws IOException
	 */
	protected void genInstanceVariables(StorageSpec storageSpec) throws IOException {
		
		if (storageSpec.isSkipCodeGeneration()) return;
		
		boolean traverseChildren = true;
		if (storageSpec.isGenerateCachedAccessors()) {
			if (storageSpec instanceof GroupSpec) {
				traverseChildren = genInstanceVariable((GroupSpec)storageSpec);
			} else {
				genInstanceVariable((FieldSpec)storageSpec);			
			}
		}
		
		if (!traverseChildren) return;
		for (Iterator<ElementSpec> iter = storageSpec.getChildren().iterator(); iter.hasNext();) {
			ElementSpec spec = iter.next();
			
			// Only GroupSpecs and FieldSpecs can have ivs; never Renames or Conditions
			if (spec instanceof StorageSpec) {
				genInstanceVariables((StorageSpec)spec);
			}
		}	
	}
	
	/**
	 * Generate a <code>gen&lt;FieldName&gt;_Endoffset()</code> method used
	 * to calculate the ending offset of a ODO field.
	 * @param storageSpec
	 * @throws IOException
	 */
	protected void genOdoEndOffsetMethod(StorageSpec storageSpec) throws IOException {
		// protected int get<javaName>_EndOffset() {
		//		int start = <fieldName>.getOffset();
		//		or, for VariableOffset cases:
		//		int start = get<prevOdoJavaName>_EndOffset() + <fieldName>.getOffset();
		//		int curLen = get<OdoObject>() * <fieldName>.getByteLength();
		//      return start + curLen;
		// }
		String javaName = getInstanceJavaName(storageSpec, true);
		String fieldName = getStaticJavaName(storageSpec);
		OccursDependingOn odo = (OccursDependingOn)storageSpec.getOccurs();
		StorageSpec odoObject = odo.getOdoObject();
		Offset offset = storageSpec.getOffset();

		formatter.println();
		formatter.println("protected int get%_EndOffset() {", javaName);
		formatter.print("int start = ");
		if (storageSpec.isVarLoc()) {
			VariableOffset vOffset = (VariableOffset)offset;
			StorageSpec previousOdo = vOffset.getOdo();
			formatter.print("get%_EndOffset() + ", 
					getInstanceJavaName(previousOdo, true));
		}
		if (storageSpec instanceof GroupSpec && storageSpec.isArray()) {
			formatter.println("%_offset;", fieldName);
		} else {
			formatter.println("%.getOffset();", fieldName);
		}
		
		formatter.println("int curLen = get%() * %%",
				getInstanceJavaName(odoObject, true),
				fieldName,
				(storageSpec instanceof GroupSpec && storageSpec.isArray())
					? "_len;"
					: ".getByteLength();");
		formatter.println("return start + curLen;");
		formatter.println("}");
	}

	/**
	 * Generate an initializer for an ODO object (size) field.
	 * @param fieldSpec
	 * @throws IOException
	 */
	protected void genOdoObjectInitializer(FieldSpec fieldSpec) throws IOException {

		if (!fieldSpec.containsOdoObject()) return;
		String ivName = getInstanceJavaName(fieldSpec, false);
		String fieldName = getStaticJavaName(fieldSpec);
		String wrapperClass;
		String fieldGetter;
		if (fieldSpec.getField() instanceof LongAccessor) {
			wrapperClass = "Long";
			fieldGetter = "getLong";
		} else {
			wrapperClass = "Integer";
			fieldGetter = "getInt";
		}
		
		formatter.println("% = new %(%.%(_byteBuffer%));",
				ivName,
				wrapperClass,
				fieldName,
				fieldGetter,
				(useBufOffset ? ", _byteBufferOffset" : ""));
	}
	
	/**
	 * Generate the ODO initializers for a StorageSpec subtree
	 * @param storageSpec
	 * @throws IOException
	 */
	protected void genOdoObjectInitializers(StorageSpec storageSpec) throws IOException {
		if (storageSpec instanceof FieldSpec) {
			genOdoObjectInitializer((FieldSpec)storageSpec);
			return;
		} else if (storageSpec instanceof GroupSpec) {
			for (Iterator<ElementSpec> iter = storageSpec.getChildren().iterator(); iter.hasNext();) {
				ElementSpec spec = iter.next();
				if (spec instanceof StorageSpec) {
					genOdoObjectInitializers((StorageSpec)spec);
				}
			}
		}
	}
	
	/**
	 * Generate  <code>factory.popOffset()</code> code.
	 * @param storageSpec
	 * @throws IOException
	 */
	protected void genPopRedefinesOffset(StorageSpec storageSpec) throws IOException {
		formatter.println("static { factory.popOffset(); } ");
		factory.popOffset();
	}
	
	/**
	 * Generate  <code>factory.pushOffset()</code> code.
	 * @param storageSpec
	 * @throws IOException
	 */	
	protected void genPushRedefinesOffset(StorageSpec storageSpec) throws IOException {
		StorageSpec redefinedSpec = storageSpec.getRedefines();
		String fieldName = getStaticJavaName(redefinedSpec);
		String offset;
		if (redefinedSpec instanceof GroupSpec && !((GroupSpec)redefinedSpec).isGenerateStringField()) {
			offset = fieldName + "_offset";
		} else {
			offset = fieldName + ".getOffset()";
		}
		formatter.println("static { factory.pushOffset(); } ");
		formatter.println("static { factory.setOffset(%); } ",
				offset);
		factory.pushOffset();
		factory.setOffset(redefinedSpec.getOffset().getOffset());
	}
	

	/**
	 * If there are fields that have (initial) VALUES clauses,
	 * generate a <code>setInitialValues()</code> method which initializes
	 * these fields to their VALUES.
	 * This will also generate code to initialze StringFields to blank
	 * if preInitStrings is true.
	 */
	protected void genSetInitialValuesMethod() throws IOException {
		
		if (fieldSpecsWithInitialValues.size() == 0 ) {
			return;  // a method is not necessary
		}
		formatter.println();
		formatter.println("public void setInitialValues() {");

		for (Iterator i=fieldSpecsWithInitialValues.iterator(); i.hasNext();) {
			FieldSpec fieldSpec = (FieldSpec)i.next();
			if (fieldSpec.hasNonemptyValueClause()) {
				formatter.println("set%(%_ival);",
								getInstanceJavaName(fieldSpec, true),
								getStaticJavaName(fieldSpec));
			} else if (fieldSpec.getField() instanceof StringField) {
				formatter.println("set%(\"\");",
						getInstanceJavaName(fieldSpec, true));
			}
		}
		
		formatter.println("}");
	}
	
	/**
	 * Generate comments for a spec that include the COBOL source lines
	 * that preceded it.
	 * @param spec ElementSpec
	 * @throws IOException
	 */
	protected void genSourceLineComments(ElementSpec spec) throws IOException {

		formatter.println();
		ArrayList lines = spec.getSourceLines();
		if (lines.size() == 0) {
			return; 
		}
		// To avoid printing leading or trailing blank lines,
		// we find the first non-blank line
		int ifirst = 0;
		int ilast = lines.size() -1;
		for (ifirst=0; ifirst <= ilast; ifirst++) {
			String s = (String)lines.get(ifirst);
			if (s.trim().length() > 0) {
				break;
			}
		}
		// Find the last non-blank line
		for (; ilast >= ifirst; ilast--) {
			String s = (String)lines.get(ilast);
			if (s.trim().length() > 0) {
				break;
			}
		}		
		// Print the first non-blank through last non-blank lines
		// wrapped in a "pre" comment.
		for (int i=ifirst; i<= ilast; i++) {
			if (i == ifirst) {
				formatter.println("/** <pre>");
			}
			String s = (String)lines.get(i);
			formatter.print(s);
			if (i < ilast) {
				formatter.println();
			} else {
				formatter.println(" </pre> */");
			}
		}
	}
	
	/**
	 * Generate a static constant declaration / initialization for a field given a string containing
	 * the VALUE clause found by the COBOL compiler.  Only common VALUEs clauses are supported; 
	 * others will generate a comment that indicates non-support.
	 * 
	 * @see Literal
	 * @param fieldSpec the FieldSpec which might have a VALUE clause
	 * @param fieldName the name of the field
	 * @param valueClause the value of the VALUE clause from the SYSADATA SYMBOL record
	 * @return true if a value was found and generated
	 * @throws IOException
	 */
	protected boolean genStaticConstantValue(FieldSpec fieldSpec, String fieldName, String valueClause) throws IOException {
		
		if (valueClause != null && valueClause.trim().length() == 0) {
			valueClause = null;  // remove an empty valueClause
		}
		if (!isPreInitialize() && valueClause == null ) {
			return false;   // we didn't generate one 
		}
		if (fieldSpec.isArray()) {
			formatter.println("/** NOTE: constant values not supported for array fields */");
			return false;
		}
		Field field = fieldSpec.getField();
		String constExpr = null;
		if (valueClause != null) {
			constExpr = Literal.toJavaCode(
								valueClause,
								field);
			if (constExpr == null) {
				formatter.println("/** NOTE: constant values not supported for this field */");
				return false;
			}			
		}
		
		String javaType;
		String preInitValue;
		if (field instanceof IntAccessor) {
			javaType = "int";
			preInitValue = "ZERO";
		} else if (field instanceof LongAccessor) {
			javaType = "long";
			preInitValue = "ZERO";
		} else if (field instanceof BigDecimalAccessor) {
			javaType = "BigDecimal";
			preInitValue = "ZERO";
		} else if (field instanceof BigIntegerAccessor) {
			javaType = "BigInteger";
			preInitValue = "ZERO";
		} else if (field instanceof FloatAccessor) {
			javaType = "float";
			preInitValue = "ZERO";
		} else if (field instanceof DoubleAccessor) {
			javaType = "double";
			preInitValue = "ZERO";
		} else if (field instanceof StringField) {
			javaType = "String";
			preInitValue = "SPACES";
		} else {
			formatter.println("/** NOTE: constant value not supported for this field */");
			return false;
		}
		
		if (valueClause == null && isPreInitialize() && preInitValue != null) {
			// Use the preInitValue as the initial value clause
			fieldSpec.setValueClause(preInitValue);
			constExpr = Literal.toJavaCode(preInitValue, field);
		}
		
		formatter.println("public static final % % = %;",
							javaType,
							fieldName,
							constExpr);		
		return true;
	}
	
	/**
	 * Generate a static variable declaration and initialization to a Field object which
	 * is used as a field converter for the field.
	 * @param javaName the Java-fied name of the field
	 * @param field the Field object 
	 * @param precision for some Fields, the number of digits of precision
	 * @throws IOException
	 */
	protected void genStaticFieldDeclaration(String javaName, Field field, int precision) throws IOException {

		String className = getUnqualifiedClassname(field.getClass());

		if (javaName != null) {
			formatter.print("% static final % % = ",
					isGenProtectedFields() ? "protected" : "public",
					className,
					javaName);
		} else {
			formatter.print("static { ");
		}
		formatter.print("factory.get%(", className);

		// write arguments to factory "constructor", by type
		if (field instanceof StringField ) {
			formatter.print(Integer.toString(field.getByteLength()));
	
		} else if (field instanceof BinaryAsIntField) {
			BinaryAsIntField binField = (BinaryAsIntField)field;
			formatter.print(Integer.toString(precision));				
			formatter.print(", ");
			formatter.print(binField.isSigned() ? "true" : "false" );

		} else if (field instanceof BinaryAsLongField) {
			BinaryAsLongField binField = (BinaryAsLongField)field;
			formatter.print(Integer.toString(precision));				
			formatter.print(", ");
			formatter.print(binField.isSigned() ? "true" : "false" );

		} else if (field instanceof BinaryAsBigIntegerField) {
			BinaryAsBigIntegerField binField = (BinaryAsBigIntegerField)field;
			formatter.print(Integer.toString(precision));				
			formatter.print(", ");
			formatter.print(Integer.toString(binField.getScale()));
			formatter.print(", ");
			formatter.print(binField.isSigned() ? "true" : "false" );
			
		} else if (field instanceof BinaryAsBigDecimalField) {
			BinaryAsBigDecimalField binField = (BinaryAsBigDecimalField)field;
			formatter.print(Integer.toString(precision));				
			formatter.print(", ");
			formatter.print(Integer.toString(binField.getScale()));
			formatter.print(", ");
			formatter.print(binField.isSigned() ? "true" : "false" );
			
		} else if (field instanceof PackedDecimalAsIntField) {
			PackedDecimalAsIntField pdField = (PackedDecimalAsIntField)field;
			formatter.print(Integer.toString(precision));				
			formatter.print(", ");
			formatter.print(pdField.isSigned() ? "true" : "false" );

		} else if (field instanceof PackedDecimalAsLongField) {
			PackedDecimalAsLongField pdField = (PackedDecimalAsLongField)field;
			formatter.print(Integer.toString(precision));				
			formatter.print(", ");
			formatter.print(pdField.isSigned() ? "true" : "false" );

		} else if (field instanceof PackedDecimalAsBigIntegerField) {
			PackedDecimalAsBigIntegerField pdField = (PackedDecimalAsBigIntegerField)field;
			formatter.print(Integer.toString(precision));				
			formatter.print(", ");
			formatter.print(Integer.toString(pdField.getScale()));
			formatter.print(", ");
			formatter.print(pdField.isSigned() ? "true" : "false" );

		} else if (field instanceof PackedDecimalAsBigDecimalField) {
			PackedDecimalAsBigDecimalField pdField = (PackedDecimalAsBigDecimalField)field;
			formatter.print(Integer.toString(precision));				
			formatter.print(", ");
			formatter.print(Integer.toString(pdField.getScale()));
			formatter.print(", ");
			formatter.print(pdField.isSigned() ? "true" : "false" );

		} else if (field instanceof ExternalDecimalAsIntField) {
			ExternalDecimalAsIntField edField = (ExternalDecimalAsIntField)field;
			formatter.print(Integer.toString(precision));				
			formatter.print(", ");
			formatter.print(edField.isSigned() ? "true" : "false" );
			formatter.print(", ");
			formatter.print(edField.isSignTrailing() ? "true" : "false");
			formatter.print(", ");
			formatter.print(edField.isSignExternal() ? "true" : "false");
			formatter.print(", ");
			formatter.print(edField.isBlankWhenZero() ? "true" : "false");

		} else if (field instanceof ExternalDecimalAsLongField) {
			ExternalDecimalAsLongField edField = (ExternalDecimalAsLongField)field;
			formatter.print(Integer.toString(precision));				
			formatter.print(", ");
			formatter.print(edField.isSigned() ? "true" : "false" );
			formatter.print(", ");
			formatter.print(edField.isSignTrailing() ? "true" : "false");
			formatter.print(", ");
			formatter.print(edField.isSignExternal() ? "true" : "false");
			formatter.print(", ");
			formatter.print(edField.isBlankWhenZero() ? "true" : "false");
			
		} else if (field instanceof ExternalDecimalAsBigIntegerField) {
			ExternalDecimalAsBigIntegerField edField = (ExternalDecimalAsBigIntegerField)field;
			formatter.print(Integer.toString(precision));				
			formatter.print(", ");
			formatter.print(Integer.toString(edField.getScale()));
			formatter.print(", ");
			formatter.print(edField.isSigned() ? "true" : "false" );
			formatter.print(", ");
			formatter.print(edField.isSignTrailing() ? "true" : "false");
			formatter.print(", ");
			formatter.print(edField.isSignExternal() ? "true" : "false");
			formatter.print(", ");
			formatter.print(edField.isBlankWhenZero() ? "true" : "false");
			
		} else if (field instanceof ExternalDecimalAsBigDecimalField) {
			ExternalDecimalAsBigDecimalField edField = (ExternalDecimalAsBigDecimalField)field;
			formatter.print(Integer.toString(precision));				
			formatter.print(", ");
			formatter.print(Integer.toString(edField.getScale()));
			formatter.print(", ");
			formatter.print(edField.isSigned() ? "true" : "false" );
			formatter.print(", ");
			formatter.print(edField.isSignTrailing() ? "true" : "false");
			formatter.print(", ");
			formatter.print(edField.isSignExternal() ? "true" : "false");
			formatter.print(", ");
			formatter.print(edField.isBlankWhenZero() ? "true" : "false");
			
		} else if (field instanceof ExternalFloatField) {
			ExternalFloatField fddField = (ExternalFloatField)field;
			formatter.print(Integer.toString(precision));				
			formatter.print(", ");
			formatter.print(Integer.toString(fddField.getScale()));
			formatter.print(", ");
			formatter.print(fddField.isShowMantissaPlusSign() ? "true" : "false" );
			formatter.print(", ");
			formatter.print(fddField.isShowExponentPlusSign() ? "true" : "false" );
			formatter.print(", ");
			formatter.print(fddField.isImpliedDecimal() ? "true" : "false" );

		} else if (field instanceof FloatAccessor || field instanceof DoubleAccessor) {
			// no factory constructor args required
			
		} else {  // the rest
			formatter.print(Integer.toString(field.getByteLength()));
		}

		if (javaName != null) {
			formatter.println(");");
		} else {
			formatter.println("); } ");
		}

	}

	/**
	 * Generate the static variables to contain Field (converter) objects which map the
	 * layout of the COBOL record.
	 * @throws IOException
	 */
	protected void genStaticFieldLayout() throws IOException {
		
		String fcn = getUnqualifiedClassname(factory.getClass());
		formatter.println("protected static % factory = new %();",
					fcn,
					fcn);
		formatter.println("static { factory.setStringTrimDefault(%); } ",
				getStringTrimDefault() ? "true" : "false");		
		if (getStringEncoding() != null) {
			formatter.println("static { factory.setStringEncoding(\"%\"); } ",
					getStringEncoding());					
		}
		genStaticFieldLayout(specTree);
	}
	
	/** 
	 * During the generation of the static field layout, we generate
	 * constants for the values of any ConditionSpecs
	 */
	protected void genStaticFieldLayout(ConditionSpec conditionSpec) throws IOException {
	
		genSourceLineComments(conditionSpec);
		if (!(conditionSpec.getParent() instanceof FieldSpec)) {
			formatter.println("/** NOTE: Group conditions are not supported */");
			return;  // we only support FieldSpecs
		}
	
		boolean supported = genStaticConstantValue((FieldSpec)conditionSpec.getParent(),
								getStaticJavaName(conditionSpec),
								conditionSpec.getValueClause());
		if (!supported) {
			conditionSpec.setValueClause(null);
		}
	}

	/**
	 * Generate a static variable declaration and initialization for a Field
	 * @param fieldSpec the FieldSpec
	 * @throws IOException
	 */
	protected void genStaticFieldLayout(FieldSpec fieldSpec) throws IOException {

		// Adjust the current factory offset if it doesn't match the field offset
		// The appearance of these in the code is normally a problem/bug
		Offset offset = fieldSpec.getOffset();
		int fieldOffset = offset.getOffset();
		if (factory.getOffset() != fieldOffset) {
			formatter.println("static { factory.incrementOffset(%); } ",
					Integer.toString(fieldOffset - factory.getOffset()));
		}
		factory.setOffset(fieldOffset);
		
		Field field = fieldSpec.getField();
		
		String javaName = getStaticJavaName(fieldSpec);
		genStaticFieldDeclaration(javaName, field, fieldSpec.getPrecision());		
			
		// If field has an occurs clause, then reserve space for the array
		// If the field has an ODO clause, then we reserve space for the max size
		Occurs occurs = fieldSpec.getOccurs();
		if (occurs != null) {

			String nen = Integer.toString(occurs.getMax());
			if (javaName != null) {
				nen = javaName + "_num";
				formatter.println("public static final int % = %;",
						nen,
						Integer.toString(occurs.getMax()));
			}
			
			if (occurs instanceof OccursDependingOn) {
				formatter.println("static { factory.setOffset(0); } // storage following is variably located");
				factory.setOffset(0);
				
			} else {
				formatter.println("static { factory.incrementOffset((% - 1) * %); } ",
						nen,
						javaName != null 
							? javaName + ".getByteLength()" 
							: Integer.toString(field.getByteLength()));
				factory.incrementOffset(occurs.getMax() * field.getByteLength());
			}

		} else {
			factory.setOffset(factory.getOffset() + field.getByteLength());
		}
		
		// For non-Array fieldSpecs with "VALUE" clause, we generate a static constant
		// for the initial value.  
		boolean found = genStaticConstantValue(fieldSpec,
								getStaticJavaName(fieldSpec) + "_ival",
								fieldSpec.getValueClause());
		
		// If we actually generate an initial value constant, 
		// then remember this FieldSpec
		if (found) {
			fieldSpecsWithInitialValues.add(fieldSpec);
		}
		
	}

	/**
	 * Generated static field layout code for a GroupSpec
	 * @param groupSpec
	 * @return boolean true if children of this class should be traversed
	 * @throws IOException
	 */
	protected boolean genStaticFieldLayout(GroupSpec groupSpec) throws IOException {
		
		int fieldOffset = groupSpec.getOffset().getOffset();
		String javaName = getStaticJavaName(groupSpec);

		// public static final int <fieldName>_len = <len>;
		int sz = groupSpec.getSize();
		formatter.println("public static final int %_len = %; %",
				javaName,
				Integer.toString(sz),
				sz < 0 ? "// contains ODO" : "");
					
		// If the group has an OCCURS or ODO clause stop traversing as it will be
		// generated as an inner class.
		if (groupSpec.isArray()) {
		
			int maxnum = groupSpec.getOccurs().getMax();
			int size = groupSpec.getSize();
			
			// Write out the fixed portion of the offset so that it can
			// be referenced later by the varloc fields.
			formatter.println("public static final int %_offset = factory.getOffset();",
					javaName);

			String nen = javaName + "_num";
			formatter.println("public static final int % = %;",
					nen,
					Integer.toString(maxnum));

			if (groupSpec.isOdo()) {
				// Reset the factory offset to 0 - what follows are variably located
				// items.
				formatter.println("static { factory.setOffset(0); } // storage following is variably located");
				factory.setOffset(0);
			} else {
				String innerClassName = getInstanceJavaName(groupSpec, true);  
				formatter.println("static { factory.incrementOffset(% * %.%_len); } ",
						nen,
						innerClassName,
						javaName);
				factory.incrementOffset(maxnum * size);
			}
			
			// The entire subtree under a Group Array will have fixed storage locations
			// relative to the inner class instance that we generate.
			removeVariableOffsetsFromAllChildren(groupSpec, groupSpec.getOffset().getOffset());
			
			return false;  // -> don't traverse children
			
		// If not an array, then we might need to generate a StringField overlay	
		} else if (groupSpec.isGenerateStringField()) {
			
			StringField field = new StringField(fieldOffset, groupSpec.getSize(), false);
			genStaticFieldDeclaration(javaName, field, 0);
				
			// reset the offset to the beginning of the group
			formatter.println("static { factory.setOffset(%.getOffset()); } ",
					javaName);

			return true; // -> continue traversing children
		
		// If a simple group, generate an offset so that it can be referenced later 
		// by a REDEFINEs clause.  Do not generate for the top level group in the 
		// class (or inner class) as this element cannot be REDEFINEd.
		} else if (groupSpec != specTree) {
			
			// Write out the fixed portion of the offset so that it can
			// be referenced later if needed by redefines
			formatter.println("public static final int %_offset = factory.getOffset();",
					javaName);

			return true; // -> continue traversing children
		} else {
			return true; // -> continue traversing children
		}
	}
	
	/**
	 * Generate the static Field declarations for a StorageSpec (sub)tree.
	 * @param storageSpec
	 * @throws IOException
	 */
	protected void genStaticFieldLayout(StorageSpec storageSpec) throws IOException {

		genSourceLineComments(storageSpec);
		
		if (genDebugComments) {
		formatter.println("//DEBUG % offset=% , factory offset=% (0x%)",
							storageSpec.getSymbol(),
							storageSpec.getOffset(),
							Integer.toString(factory.getOffset()),
							Integer.toString(factory.getOffset(),16));
		}
		
		if (storageSpec.isSkipCodeGeneration()) return;
		
		if (storageSpec.isRedefines()) { 
			genPushRedefinesOffset(storageSpec);
		}
		boolean traverseChildren = true;
		if (storageSpec instanceof GroupSpec) {
			traverseChildren = genStaticFieldLayout((GroupSpec)storageSpec);
		} else {
			genStaticFieldLayout((FieldSpec)storageSpec);
		}
		
		if (!traverseChildren) return;
		for (Iterator<ElementSpec> iter = storageSpec.getChildren().iterator(); iter.hasNext();) {
			ElementSpec spec = iter.next();
			if (spec instanceof StorageSpec) {
				genStaticFieldLayout((StorageSpec)spec);
			} else if (spec instanceof ConditionSpec) {
				genStaticFieldLayout((ConditionSpec)spec);
			} else {
				// RENAMES are not currently supported
				genSourceLineComments(spec);
				formatter.println("// NOTE: Unsupported COBOL element feature");
			}
		}
		if (storageSpec.isRedefines()) {
			genPopRedefinesOffset(storageSpec);
		}
	}
	

	/**
	 * Generate a Java instance-level name for a COBOL data name in the given ElementSpec.
	 * A list of "Name_In_Name"s are generated until the last Name segement
	 * is unique in the Copy book.
	 * 
	 * @see #getInstanceJavaName(String, boolean)
	 * @param spec the ElementSpec
	 * @param upperFirst if true, the first letter is uppercased
	 * @return String
	 */
	protected String getInstanceJavaName(ElementSpec spec, boolean upperFirst ) {
		StringBuffer sb = new StringBuffer();
		for (;;) {
			sb.append(getInstanceJavaName(spec.getSymbol(), upperFirst));
			if (spec.isSymbolUnique()) break;
			sb.append("_In_");
			spec = spec.getParent();
			if (spec == null) break;
		}
		return sb.toString();
	}
	
	/**
	 * Convert a COBOL data name symbol to a Java name.  This is done
	 * by Uppercasing the first letter of each COBOL-dash-delimited word
	 * @param symbol the COBOL data name
	 * @param upperFirst true if the first word is uppercased
	 * @return String
	 */
	protected String getInstanceJavaName(String symbol, boolean upperFirst ) {
		StringTokenizer strtok = new StringTokenizer(symbol, "-");
		StringBuffer sb = new StringBuffer();
		while (strtok.hasMoreTokens()) {
			String tok = strtok.nextToken();
			if (!upperFirst) {
				sb.append(tok.toLowerCase());
				upperFirst = true;
			} else {
				sb.append(Character.toUpperCase(tok.charAt(0)));
				sb.append(tok.substring(1).toLowerCase());
			}
		}
		return sb.toString();
	}
	
	/**
	 * Generate a Java static-level name for a COBOL data name in the given ElementSpec.
	 * A list of "NAME_In_NAME"s are generated until the last Name segement
	 * is unique in the Copy book.
	 * 
	 * @see #getStaticJavaName(String)
	 * @param spec the ElementSpec
	 * @return String
	 */
	protected String getStaticJavaName(ElementSpec spec ) {
		StringBuffer sb = new StringBuffer();
		for (;;) {
			sb.append(getStaticJavaName(spec.getSymbol()));
			if (spec.isSymbolUnique()) break;
			sb.append("_In_");
			spec = spec.getParent();
			if (spec == null) break;
		}
		return sb.toString();
	}

	/**
	 * Convert a COBOL data name symbol to a Java name.  This is done
	 * by Uppercasing and underscore delimiting each COBOL-dash-delimited word
	 * @param symbol the COBOL data name
	 * @return String
	 */
	protected String getStaticJavaName(String symbol ) {
		StringTokenizer strtok = new StringTokenizer(symbol, "-");
		StringBuffer sb = new StringBuffer();
		while (strtok.hasMoreTokens()) {
			String tok = strtok.nextToken();
			sb.append(tok.toUpperCase());
			if (strtok.hasMoreTokens()) {
				sb.append("_");
			}
		}
		return sb.toString();
	}

	/**
	 * The entire subtree under a Group Array will have fixed storage locations
	 * relative to the inner class instance that we generate.
	 */
	protected void removeVariableOffsetsFromAllChildren(GroupSpec groupSpec, int baseOffset) {
		
		for (Iterator<ElementSpec> i=groupSpec.getChildren().iterator(); i.hasNext();) {
			ElementSpec spec = i.next();
			if (spec instanceof StorageSpec) {
				spec.setOffset(new Offset(spec.getOffset().getOffset() - baseOffset));
				spec.setVarLoc(false);
			}
			if (spec instanceof GroupSpec) {
				removeVariableOffsetsFromAllChildren((GroupSpec)spec, baseOffset);
			}
		}
	}

	/**
	 * Check that we have a specTree and this it is a GROUP without an OCCURS clause
	 */
	protected void validateSpecTree() {
		if (specTree == null) {
			throw new IllegalArgumentException("spec tree not given");
		}
		if (!(specTree instanceof GroupSpec)) {
			throw new IllegalArgumentException("top-level element must be a GROUP");
		}
		if (specTree.isArray()) {
			throw new IllegalArgumentException("top-level GROUP with OCCURS clause is not supported");
		}
	}

}
