/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007, 2012. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.cobol;

import java.io.BufferedWriter;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import com.ibm.jzos.fields.CobolDatatypeFactory;
import com.ibm.jzos.fields.Field;
import com.ibm.jzos.recordgen.ADataRecord;
import com.ibm.jzos.recordgen.ADataRecordReader;

/**
 * A class that converts a COBOL Compiler AData file into an abstract
 * {@link ElementSpec} object model tree.
 * <p>
 * This class is used by {@link RecordClassGenerator} to generate Java source
 * code from COBOL AData files.
 * <p/>
 * Note: this class is not intended to be a public interface;
 * please use {@link RecordClassGenerator}
 * <p>
 * @see ADataRecord
 * @see ADataSourceRecord
 * @see ADataSymbolRecord
 * @see ElementSpec
 */
public class RecordSpecTreeBuilder {

	/**
	 * Inner class used to keep track of the running field offset in the ElementSpec
	 * tree.  This class is used once an Occurs Depending On (ODO) clause is encountered
	 * in a copybook.  All subsequent fields have no fixed offset from the beginning of
	 * the record, so the offsets must be calculated on the runtime size of the ODO
	 * group (or field). 
	 */
	private static class RunningVariableOffset {
		private ArrayList<VariableOffset> offsets = new ArrayList<VariableOffset>();
		public VariableOffset getOffset() {
			return (VariableOffset)basicGet().clone();
		}
		public void incrementOffset(int increment) {
			VariableOffset o = basicGet();
			o.incrementOffset(increment);
		}
		public void pop() {
			offsets.remove(offsets.size() - 1);
		}
		public void push(VariableOffset offset) {
			offsets.add((VariableOffset)offset.clone());
		}
		public void setOffset(VariableOffset offset) {
			offsets.add(offset);
		}
		private VariableOffset basicGet() {
			return offsets.get(offsets.size()-1);
		}
	};
	
	/**
	 * Test method that reads a COBOL ADATA file and produces a spec tree.
	 * @param args The first argument (required) is the ADATA file.  The second argument
	 * (optional) is the symbol to use as the root of the tree.  If no symbol name is
	 * supplied, the first group symbol encountered in the ADATA file will be used.
	 * @throws Exception
	 */
	public static void main(String[] args) throws Exception {
		if (args.length < 1) {
			System.out.println("USAGE: SpecTreeBuilder <adata_filename> [symbol]");
			System.exit(1);
		}
		RecordSpecTreeBuilder specBuilder = new RecordSpecTreeBuilder();
		if (args.length == 2) {
			specBuilder.setSelectedSymbolName(args[1]);
		}
		ElementSpec specTree = specBuilder.process(args[0]);

		//dump spec tree
		BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(System.out));
		specTree.dumpOn(writer);
		writer.close();
	}
	
	/** Writer used for debugging */
	BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(System.err));
	
	/** If true, all COBOL ADATA source records are dumped to stderr */
	private boolean dumpSourceRecords = false;
	
	/** CobolDatatypeFactory used to generate the {@link Field} objects associated with {@link FieldSpec}s. */
	private CobolDatatypeFactory factory = new CobolDatatypeFactory();

	/** Used to assign indexed names to "FILLER" data items */
	private int fillerIndex = 1;
	
	/** Back pointer to the last source line consumed.  A starting point for the current ElementSpec */
	private int previousSourceIndex = 0;
	
	/** Used to keep track of the running offset in records that have an Occurs Depending On element*/
	private RunningVariableOffset runningVariableOffset = new RunningVariableOffset();	
	
	/** The symbol used to build the tree.  If null, the first group symbol in the ADATA file is used */
	private String selectedSymbolName = null;
	
	/** List of COBOL ADATA source records read.  Used to associate source lines with Specs */
	private ArrayList sourceRecords = new ArrayList();
	
	/** Map used to lookup spec parentage and association */
	private Map<Integer,ElementSpec> specsById = new HashMap<Integer,ElementSpec>();

	/** Root of the RecordSpec tree */
	private ElementSpec specTree = null;
	
	/**
	 * The symbol name to begin the record with.
	 * @return String the selected symbol name
	 */
	public String getSelectedSymbolName() {
		return selectedSymbolName;
	}

	/**
	 * Process a COBOL AData file.
	 * @param filename the name of the ADATA file.  May be either a file path or
	 * a dataset name (beginning with //).
	 * <p/>
	 * Note: this class is not intended to be a public interface;
	 * please use {@link RecordClassGenerator}
	 * <p/>
	 * @return ElementSpec the resulting spec tree
	 * @throws IOException
	 */
	public StorageSpec process(String filename) throws IOException {
		
		ADataSymbolRecord startRec = null;
		ADataRecordReader reader = new ADataRecordReader(filename);
		sourceRecords.add("dummy 0th element");
		ADataRecord rec = null;
		while ((rec = reader.readRecord()) != null) {
			if (rec.getRECTYPE() == ADataSourceRecord.ADATA_RECSOURCE) {
				sourceRecords.add(new ADataSourceRecord(rec.getBytes()));
				if (dumpSourceRecords) {
					rec.dumpOn(writer);
					writer.flush();
				}
			} else if (rec.getRECTYPE() == ADataSymbolRecord.ADATA_RECSYM) {
				ADataSymbolRecord srec = new ADataSymbolRecord(rec.getBytes());
				if (startRec == null) {
					if ((selectedSymbolName == null && srec.isGroup()) ||
					    (srec.getSymbolName().equalsIgnoreCase(selectedSymbolName))) 
					{
						startRec = srec;
						processSymbolRecord(srec);
					}
				} else if (srec.getLevel() >= startRec.getLevel()) {
					processSymbolRecord(srec);
				}
			}
		}
			
		// Calculate the variable offsets in cases where ODO clauses occur.
		if (!(specTree instanceof StorageSpec)) {
			throw new IllegalStateException("specTree root not a StorageSpec");
		}
		calculateVariableOffsets(null, specTree);
		calculateStorageOverlays(specTree, true);
		return (StorageSpec)specTree;
	}

	/**
	 * Sets a debug flag which causes the AData source records to be dumped to STDERR
	 * @param dumpSourceRecords
	 */
	public void setDumpSourceRecords(boolean dumpSourceRecords) {
		this.dumpSourceRecords = dumpSourceRecords;
	}
	
	/**
	 * @see #getSelectedSymbolName()
	 */
	public void setSelectedSymbolName(String selectedSymbolName) {
		this.selectedSymbolName = selectedSymbolName;
	}
	
	/**
	 * Return the spec given its unique id
	 */
	public ElementSpec getSpecWithId(int id) {
		return (ElementSpec)specsById.get(new Integer(id));
	}
	
	/**
	 * Build a ConditionSpec to represent a level 88 COBOL copybook element.
	 * @param symRec
	 * @return ConditionSpec
	 */
	private ConditionSpec buildConditionSpec(ADataSymbolRecord symRec) {
		ConditionSpec conditionSpec = new ConditionSpec(symRec.getValueClause());
		return conditionSpec;
	}

	/**
	 * Build a FieldSpec from symRec.  The resulting FieldSpec will contain an
	 * appropriate {@link Field} subclass.
	 * @param symRec
	 * @return FieldSpec
	 */
	private FieldSpec buildFieldSpec(ADataSymbolRecord symRec) {
		FieldSpec fieldSpec = new FieldSpec();
		fieldSpec.setField(getField(symRec));
		fieldSpec.setComp5(symRec.isComp5());
		fieldSpec.setValueClause(symRec.getValueClause());
		if (symRec.isNumeric()) {
			 // need to save precision since Binary*Fields don't have it anywhere
			fieldSpec.setPrecision(symRec.getPrecision()); 
		}
		return fieldSpec;
	}

	/**
	 * Build a spec for a container (group) element.
	 * @param symRec
	 * @return GroupSpec
	 */
	private GroupSpec buildGroupSpec(ADataSymbolRecord symRec) {
		GroupSpec groupSpec = new GroupSpec();
		groupSpec.setValueClause(symRec.getValueClause());
		return groupSpec;
	}
	
	/**
	 * Build a RenameSpec to represent a level 66 COBOL copybook element.
	 * @param symRec
	 * @return RenameSpec
	 */
	private RenameSpec buildRenameSpec(ADataSymbolRecord symRec) {
		RenameSpec renameSpec = null;
		int startRenamedId = symRec.getStrtRenamedId();
		StorageSpec fromStorageSpec = (StorageSpec)specsById.get(new Integer(startRenamedId));
		if (fromStorageSpec == null) {
			throw new IllegalStateException("Null fromStorageSpec (id=" + startRenamedId + ") in " + symRec.getSymbolName());
		}
		int endRenamedId = symRec.getEndRenamedId();
		if (endRenamedId == 0) {
			//Single StorageSpec rename
			renameSpec = new RenameSpec(fromStorageSpec);
		} else {
			StorageSpec toStorageSpec = (StorageSpec)specsById.get(new Integer(endRenamedId));
			if (toStorageSpec == null) {
				throw new IllegalStateException("Null toStorageSpec (id=" + endRenamedId + ") in " + symRec.getSymbolName());
			}
			renameSpec = new RenameRangeSpec(fromStorageSpec, toStorageSpec);
		}
		return renameSpec;
	}

	/**
	 * Traverse tree in Depth First order and mark any StorageSpecs that might
	 * share storage with other StorageSpecs as "generateCachedAccessors=false".
	 * <p>
	 * We only worry about marking collateral damage resulting from "redefines", 
	 * as collateral GroupSpec overlay StringFields will be explicitly managed.
	*/
	static void calculateStorageOverlays(ElementSpec elementSpec, boolean addWarnings) {
		if (!(elementSpec instanceof StorageSpec)) {
			return;
		}
		StorageSpec spec = (StorageSpec)elementSpec;
		if (spec.isRedefines()) {
			StorageSpec redefinedSpec = spec.getRedefines();
			if (redefinedSpec.containsOdoObject()) {
				spec.setSkipCodeGeneration(true);
				if (addWarnings) {
					spec.getSourceLines()
						.add("*** NOTE: REDEFINE of ODO object is not supported; ignored" );
				}
			} else {
				spec.setGenerateCachedAccessors(false);
				redefinedSpec.setGenerateCachedAccessors(false);
			}
		}
		for (Iterator<ElementSpec> iter = spec.getChildren().iterator(); iter.hasNext();) {
			calculateStorageOverlays(iter.next(), addWarnings);
		}		
	}
	
	/**
	 * Traverse tree in Depth First order to create variable offsets, if needed.
	 * <p>
	 * Processing:
	 * <ul>
	 *   <li>If Spec is not variably located (ODO has not been seen):</li>
	 *     <ul>
	 *     <li>Keep fixed offset that was created from the ADataSymbolRecord</li>
	 *     <li>Visit children</li>
	 *     </ul>
	 *   <li>If spec is variably located (ODO has been seen):</li>
	 *     <ul>
	 *     <li>Save <strong>copy</strong> of running VariableOffset in Spec</li>
	 *     <li>Visit children</li>
	 *     <li>If FieldSpec, add size to running VariableOffset.  GroupSpec sizes are ignored</li>
	 *     <li>Visit children</li>
	 *     </ul>
	 *     <p>If this spec has a REDEFINES clause, the running VariableOffset is pushed on
	 *     a Stack, leaving causes it to be popped.  This enables redefines to be handled
	 *     without affecting the overall storage map.
	 * </ul>
	*/
	static void calculateVariableOffsets(RunningVariableOffset runningVariableOffset, ElementSpec elementSpec) {
		if (!(elementSpec instanceof StorageSpec)) {
			return;
		}
		if (runningVariableOffset == null) {
			runningVariableOffset = new RunningVariableOffset();  // lazy initialize at top StorageSpec
		}
		StorageSpec spec = (StorageSpec)elementSpec;
		if (!spec.isVarLoc()) {
			for (Iterator<ElementSpec> iter = spec.getChildren().iterator(); iter.hasNext();) {
				calculateVariableOffsets(runningVariableOffset, iter.next());
			}			
			if (spec.isOdo()) {
				runningVariableOffset.setOffset(new VariableOffset(spec));
			}
		} else {
			if (spec.isRedefines()) {
				runningVariableOffset.push((VariableOffset)spec.getRedefines().getOffset());
			}
			spec.setOffset(runningVariableOffset.getOffset());  // HERE is what we actually do!
			for (Iterator<ElementSpec> iter = spec.getChildren().iterator(); iter.hasNext();) {
				calculateVariableOffsets(runningVariableOffset, iter.next());
			}
			if (spec instanceof FieldSpec) {
				runningVariableOffset.incrementOffset(spec.getSize());
			}
			if (spec.isOdo()) {
				runningVariableOffset.setOffset(new VariableOffset(spec));
			}
			if (spec.isRedefines()) {
				runningVariableOffset.pop();
			}
		}
	}
	
	/**
	 * Return a subclass of {@link Field} that matches the attributes in symRec.
	 * @param symRec
	 * @return Field
	 */
	private Field getField(ADataSymbolRecord symRec) {
		Field field = null;
		if (symRec.isUsageNationalOrDBCS()) {
			// We currently don't support NATIONAL or DBCS (DISPLAY-1) fields 
			field = factory.getByteArrayField(symRec.getSize());			
		} else if (symRec.isNumeric()) {
			field = getNumericField(symRec);
		} else if (symRec.isCharacter()) {
			field = factory.getStringField(symRec.getSize(), symRec.getSize() > 1, symRec.isJustified());
		} else {
			field = factory.getByteArrayField(symRec.getSize());			
		}
		//Set the field offset to the structure offset.  During code
		//generation, special handling will be required for variable located items.
		field.setOffset(symRec.getStructOffset());
		return field;
	}
	
	/**
	 * Return a numeric {@link Field} that matches the attributes in symRec.
	 * @param symRec
	 * @return Field
	 */
	private Field getNumericField(ADataSymbolRecord symRec) {
		Field field = null;
		int scale = symRec.isScaledNegative() ? -symRec.getScale() : symRec.getScale();
		if (symRec.isNumericNational()) {
			// Nationals are not supported; we map to byte arrays
			field = factory.getByteArrayField(symRec.getSize());
		} else if (symRec.isBinary()) {
			field = factory.getBinaryField(symRec.getPrecision(), scale, symRec.isSigned(), symRec.isComp5());
		} else if (symRec.isExternalFloatingPoint()) {
			field = factory.getExternalFloatField(symRec.getPrecision(), scale, symRec.isShowMantissaPlusSign(), symRec.isShowExponentPlusSign(), symRec.isImpliedDecimal());
		} else if (symRec.isInternalFloatingPoint()) {
			if ((symRec.getUsageClause() & ADataSymbolRecord.USAGE_COMP1) != 0) {
				field = factory.getIbmFloatField();
			} else {
				field = factory.getIbmDoubleField();				
			}
		} else if (symRec.isPackedDecimal()) {
			field = factory.getPackedDecimalField(symRec.getPrecision(), scale, symRec.isSigned());
		} else if (symRec.isNumericEdited()) {
			// If "BLANK WHEN ZERO" is the only reason why this is "numeric edited", then we handle as Ext Decimal
			// All other Numeric-edit fields we currently map to Strings
			if (symRec.isExternalDecimal() && symRec.isBlankWhenZero() && symRec.hasNumericPictureClause()) {
				field = factory.getExternalDecimalField(symRec.getPrecision(), scale, symRec.isSigned(), symRec.isSignTrailing(), symRec.isSignSeparate(), symRec.isBlankWhenZero());
			} else {
				field = factory.getStringField(symRec.getSize(), false);
			}
		} else if (symRec.isExternalDecimal()) {
			field = factory.getExternalDecimalField(symRec.getPrecision(), scale, symRec.isSigned(), symRec.isSignTrailing(), symRec.isSignSeparate(), symRec.isBlankWhenZero());
		}
		return field;
	}
	
	/**
	 * Determine which source lines are associated with the supplied symbol.
	 * The symbol's getLineNum() returns the source line containing the
	 * symbol itself, but additional source lines may be associated.  Read
	 * until a statement terminator is detected.
	 */
	private ArrayList<String> getSourceLines(ADataSymbolRecord symRec) {
		ArrayList<String> sourceLines = new ArrayList<String>();
		int sourceIndex = symRec.getLineNum();
		if (previousSourceIndex == 0) {
			previousSourceIndex = sourceIndex;
		}
		//Collect anything (comments, etc..) after the previous source line and
		//prior to the current symbol's source line.
		while (previousSourceIndex > 0 && previousSourceIndex < sourceIndex) {
			ADataSourceRecord srcRec = (ADataSourceRecord)(sourceRecords.get(previousSourceIndex++));
			sourceLines.add(getScrubbedSourceLine(srcRec));
		}
		while (true) {
			ADataSourceRecord srcRec = (ADataSourceRecord)(sourceRecords.get(sourceIndex++));
			String sourceLine = getScrubbedSourceLine(srcRec);
			sourceLines.add(sourceLine);
			//End of statement occurs when a period is the last non-whitespace
			//character in areas A&B (Columns 7 - 72).
			if (sourceLine.endsWith(".")) {
				break;
			}
		}
		previousSourceIndex = sourceIndex;
		return sourceLines;
	}
	
	/**
	 * Return a copy of the source line (Column 7 + areas A and B only), with any
	 * trailing whitespace removed, replace any java commnets and replace literal
	 * @ signs with the unicode sequence.
	 */
	private String getScrubbedSourceLine(ADataSourceRecord srcRec) {
		String sourceLine = srcRec.getSourceRecord();
		sourceLine = sourceLine.substring(6,Math.min(72, sourceLine.length()));
		int i = sourceLine.length()-1;
		//Trim trailing whitespace
		for (; i>=0 && (sourceLine.charAt(i) == ' '); i--);
		return sourceLine.substring(0,i+1);
	}

	/**
	 * Handle the OCCURS clause associated with symRec.  The simple occurs
	 * case will result in a fixed array.  The occurs depending on (ODO) case
	 * will result in a variable sized record based on the value of the ODO
	 * object.
	 * @param spec the StorageSpec
	 * @param symRec the ADATA symbol record containing the OCCURS clause.
	 */
	private void handleOccurs(StorageSpec spec, ADataSymbolRecord symRec) {
		int min = symRec.getOccursMinimum();
		int max = symRec.getOccursMaximum();
		Occurs occurs = null;
		if (symRec.isOccursDependingOnClause()) {
			configureOdoEnclosures(spec);
			FieldSpec odoObject = (FieldSpec)specsById.get(new Integer(symRec.getOdoSymbolId()));
			if (odoObject == null) {
				throw new IllegalStateException("ODO Object not found for: " + symRec.getSymbolName());
			}
			odoObject.setContainsOdoObject(true);
			occurs = new OccursDependingOn(min, max, (FieldSpec)odoObject);
		} else {
			occurs = new Occurs(max);
		}
		spec.setOccurs(occurs);	
	}

	/**
	 * Ensure that this spec has no parent with an occurs clause.
	 * Also tag each containing Group as having an ODO somewhere in its subtree.
	 */
	static void configureOdoEnclosures(StorageSpec spec) {
		for(ElementSpec p = spec.getParent(); p != null; p = p.getParent()) {
			if (!(p instanceof StorageSpec)) continue;
			StorageSpec parent = (StorageSpec)p;
			if (parent.getOccurs() != null) {
				throw new IllegalStateException(spec.getSymbol() + " is a nested ODO.");
			}
			if (parent instanceof GroupSpec) {
				((GroupSpec)parent).setContainsOdo(true);
			}
		}	
	}
	/**
	 * Handle the REDEFINES clause associated with symRec.  This clause means that this
	 * symbol redefines storage previously defined.  The end result is that redefined
	 * elements do not play a role in determining absolute record storage offsets.
	 * @param spec
	 * @param symRec
	 */
	private void handleRedefine(StorageSpec spec, ADataSymbolRecord symRec) {
		StorageSpec redefinedSpec = (StorageSpec)specsById.get(new Integer(symRec.getRedefinedId()));
		if (redefinedSpec == null) {
			throw new IllegalStateException("No redefined spec found for " + symRec.getSymbolName());
		}
		((StorageSpec)spec).setRedefines(redefinedSpec);		
	}

	/**
	 * Given a new ADataSymbolRecord (within the top level symbol), add an 
	 * appropriate ElementSpec subclass to the tree.
	 * <p>As specs are encountered, they are added to the specsById map.  This
	 * allows for later lookup of spec dependencies such as parentage, redefines,
	 * ODO objects, etc.
	 * @param symRec
	 */
	private void processSymbolRecord(ADataSymbolRecord symRec) {
		ElementSpec spec = null;
		if (symRec.getSymbolType() != ADataSymbolRecord.SYMT_DATA_NAME) {
			return;
		}
		if (symRec.getStrtRenamedId() != 0) {
			spec = buildRenameSpec(symRec);
	    } else if (symRec.isGroup()) {
			spec = buildGroupSpec(symRec);
		} else if (symRec.isField()) {
			spec = buildFieldSpec(symRec);
		} else if (symRec.isCondition()) {
			spec = buildConditionSpec(symRec);
		} else {
			throw new IllegalArgumentException("Illegal or unsupported symbol record");
		}
		if (symRec.getFlags1() == ADataSymbolRecord.FLG1_FILLER) {
			spec.setSymbol("FILLER_" + fillerIndex++);
			spec.setSymbolUnique(true);
		} else {
			spec.setSymbol(symRec.getSymbolName());
			spec.setSymbolUnique(symRec.getQualInd() == 0);
		}
		spec.setSourceLines(getSourceLines(symRec));
		spec.setSourceStmt(symRec.getLineNum());
		spec.setLevel(symRec.getLevel());
		// Set the structure offset contained in the symbol record
		// If this is a variable located symbol, the offset will be
		// set to a VariableOffset in calculateVariableOffsets()
		spec.setOffset(new Offset(symRec.getStructOffset()));
		// Set the spec size contained in the symbol record.  For specs with
		// OCCURS clauses, this is the size of a single element.
		spec.setSize(symRec.getSize());
		spec.setVarLoc(symRec.isVarLoc());
				
		spec.setSymbolId(symRec.getSymbolId());
		specsById.put(new Integer(spec.getSymbolId()), spec);
		int parentId = symRec.getParentId();
		if (parentId != 0) {
			StorageSpec parent = (StorageSpec) specsById.get(new Integer(parentId));
			if (parent == null) {
				throw new IllegalStateException(spec.getSymbol() + " parent not found");
			}
			spec.setParent(parent);
			parent.addChild(spec);
		}

		if (symRec.isRedefinesClause()) {
			handleRedefine((StorageSpec)spec, symRec);
		}
		
		//Now that parentage has been set, handle occurs clause, if present
		if (symRec.isOccursClause() || symRec.isOccursDependingOnClause()) {
			handleOccurs((StorageSpec)spec, symRec);
		}

		// Set the root of the tree if this is the first spec.
		if (specTree == null) {
			specTree = spec;
		}
	}

}
