/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007, 2012. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.cobol;

import java.io.IOException;
import java.io.InputStream;
import java.util.HashMap;
import java.util.Map;

import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import com.ibm.jzos.FileFactory;
import com.ibm.jzos.fields.CobolDatatypeFactory;

/**
 * Parsed the XML description created by {@link RecordXMLGenerator) and
 * returns a RecordSpec describing an Assembler DSECT or CSECT.
 * <p/>
 * Note: this class is not intended to be a public interface
 * <p/>
 * @since 2.4.4
 * @see RecordClassGenerator
 */
public class RecordSpecTreeXMLParser {

	/** The XML Stream reader **/
	private XMLStreamReader reader;

	/** Map used to lookup spec parentage and association */
	private Map<Integer, ElementSpec> specsById = new HashMap<Integer, ElementSpec>();

	/** Root of the RecordSpec tree */
	private StorageSpec specTree = null;
	
	/** A CobolDataTypeFactory used to build Field objects */
	private CobolDatatypeFactory factory = new CobolDatatypeFactory();

	/**
	 * Parse an XML file
	 * @param filename the name of the XML file
	 * @return StorageSpec (tree)
	 * @throws IOException
	 */
	public StorageSpec process(String filename) throws IOException, XMLStreamException {

		reader = makeXMLReader(filename);
		String ename;
		
		while (reader.hasNext()) {
			switch (reader.next()) {
			case XMLStreamConstants.START_ELEMENT:
				ename = reader.getLocalName();
				if (ename.equalsIgnoreCase(GroupSpec.XML_ELEMENT_NAME)) {
					specTree = new GroupSpec();
					specTree.parseXmlFrom(this);					
				} else if (ename.equalsIgnoreCase(FieldSpec.XML_ELEMENT_NAME)) {
					specTree = new FieldSpec();
					specTree.parseXmlFrom(this);					
				} else {
					throw new XMLStreamException("Unexpected element: "+ename, reader.getLocation());
				}
			}
		}
		RecordSpecTreeBuilder.calculateVariableOffsets(null, specTree);
		RecordSpecTreeBuilder.calculateStorageOverlays(specTree, false);
		return specTree;
	}
	
	private XMLStreamReader makeXMLReader(String filename) throws IOException, XMLStreamException {
		XMLStreamReader reader;
		XMLInputFactory xmlif = XMLInputFactory.newInstance();
		InputStream is = FileFactory.newInputStream(filename);
		reader = xmlif.createXMLStreamReader(is);
		return reader;
	}

	XMLStreamReader getReader() {
		return reader;
	}
	
	CobolDatatypeFactory getFactory() {
		return factory;
	}
	
	ElementSpec getSpecById(int id) {
		return specsById.get(id);
	}
	
	void addSpecById(ElementSpec spec) {
		specsById.put(spec.getSymbolId(), spec);
	}
	
	String getAttributeValue(String name) {
		return reader.getAttributeValue(null, name);
	}

	String getAttributeValue(String name, String def) {
		String v = reader.getAttributeValue(null, name);
		if (v == null) {
			return def;
		} else {
			return v;
		}
	}
	
	int getAttributeIntValue(String name) {
		return getAttributeIntValue(name, 0);
		
	}
	int getAttributeIntValue(String name, int def) {
		String v = reader.getAttributeValue(null, name);
		if (v == null) {
			return def;
		}
		return Integer.parseInt(v);
	}	
	
	boolean getAttributeBooleanValue(String name, boolean def) {
		String v = reader.getAttributeValue(null, name);
		if (v == null) {
			return def;
		}
		return "true".equalsIgnoreCase(v);
	}
	
	StorageSpec getStorageSpecForId(int id) throws XMLStreamException {
		ElementSpec e = getSpecById(id);
		if (!(e instanceof StorageSpec)) {
			throw new XMLStreamException("invalid StorageSpec id reference: "+id, 
										getReader().getLocation());
		} else {
			return (StorageSpec)e;
		}
	}
}
