/*
* %Z%%W% %I%
*
===========================================================================
* Licensed Materials - Property of IBM
* "Restricted Materials of IBM"
* (C) Copyright IBM Corp. 2007, 2012. All Rights Reserved
*
===========================================================================
*/
package com.ibm.jzos.recordgen.cobol;

import java.io.IOException;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.xml.stream.XMLStreamConstants;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;
import javax.xml.stream.XMLStreamWriter;

/**
 * Object that represents an element in a COBOL COPYBOOK that takes storage in
 * the resulting record.  This includes both GroupSpecs and FieldSpecs.
 */
public abstract class StorageSpec extends ElementSpec {
	private Occurs occurs;
	private StorageSpec redefines;
	private List<ElementSpec> children = new ArrayList<ElementSpec>();
	private String valueClause;
	private boolean generateCachedAccessors = true;
	private boolean skipCodeGeneration = false;

	
	/**
	 * Add a child element spec.  If the receiver is a GroupSpec, this will be either
	 * another (nested) GroupSpec or a FieldSpec.
	 * @param elementSpec
	 */
	public void addChild(ElementSpec elementSpec) {
		children.add(elementSpec);
	}

	/**
	 * Return the immediate children of the receiver.
	 * @return List the children
	 */
	public List<ElementSpec> getChildren() {
		return children;
	}

	/**
	 * @see #getChildren()
	 */
	public void setChildren(List<ElementSpec> children) {
		this.children = children;
	}

	/**
	 * Return the OCCURS clause information associated with the receiver.  Return
	 * null if no OCCURS clause.  If this element has an Occurs Depending On (ODO)
	 * clause, an {@link OccursDependingOn} instance is returned.
	 * @return Occurs
	 */
	public Occurs getOccurs() {
		return occurs;
	}

	/**
	 * @see #getOccurs()
	 */
	public void setOccurs(Occurs occurs) {
		this.occurs = occurs;
	}

	/**
	 * Return the StorageSpec that this spec REDEFINES.  Return null if
	 * this spec has no REDEFINES clause.
	 * @return StorageSpec the storage spec that this spec redefines
	 */
	public StorageSpec getRedefines() {
		return redefines;
	}

	/**
	 * @see #getRedefines()
	 */
	public void setRedefines(StorageSpec redefines) {
		this.redefines = redefines;
	}

	/**
	 * Return true if this spec has an OCCURS DEPENDING ON clause.
	 */
	public boolean isOdo() {
		return occurs != null && occurs instanceof OccursDependingOn;
	}
	
	/**
	 * Return true if this spec has a REDEFINES clause.
	 */
	public boolean isRedefines() {
		return redefines != null;
	}
	
	/**
	 * Return true if this spec represents an array of items
	 */
	public boolean isArray() {
		return occurs != null;
	}
	
	/**
	 * Return true if this spec contains an ODO somewhere in the subtree
	 */
	public boolean containsOdo() {
		return false;
	}
	
	/**
	 * Dump debug information for the receiver
	 * <p>
	 * @param writer
	 * @throws IOException
	 */
	public void dumpOn(Writer writer) throws IOException {

		super.dumpOn(writer);
		for (Iterator<ElementSpec> iter = getChildren().iterator(); iter.hasNext();) {
			ElementSpec spec = iter.next();
			spec.dumpOn(writer);
		}
		
	}
	
	/**
	 * A debug representation of the receiver.
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer(super.toString());
		if (redefines != null || occurs != null) {
			sb.append(" [");
			if (redefines != null) {
				sb.append("Redefines " + redefines.getSymbol());
				sb.append(" ");
			}
			if (occurs != null) {
				sb.append(occurs);
			}
			sb.append("] ");
		}
		return sb.toString();
	}

	/**
	 * Return true if instance variable caches are used for accessor methods
	 * for this StorageSpec
	 */
	public boolean isGenerateCachedAccessors() {
		return generateCachedAccessors;
	}

	/**
	 * @see #isGenerateCachedAccessors()
	 * Note: this property is initially true, and can only be set to false,
	 * and once set to false can never be set back to true.  This
	 * allows for defaults and redefine overlaps to work together.
	 */
	public void setGenerateCachedAccessors(boolean value) {
		if (!value) {
			this.generateCachedAccessors = value;
		}
	}

	/** 
	 * Answer true if this storage spec is or contains a field
	 * that is the object of a "OCCURS DEPENDS ON" clause.
	 * In other words, it contains the size of an ODO.
	 */
	public boolean containsOdoObject() {
		return false; 
	}

	/**
	 * @see #setSkipCodeGeneration(boolean)
	 */
	public boolean isSkipCodeGeneration() {
		return skipCodeGeneration;
	}

	/**
	 * Sets a flag that indicates whether code generation for this
	 * spec is skipped.  
	 */
	public void setSkipCodeGeneration(boolean skipCodeGeneration) {
		this.skipCodeGeneration = skipCodeGeneration;
	}

	public boolean hasNonemptyValueClause() {
		return valueClause != null && valueClause.trim().length() > 0;
	}
	
	/**
	 * Answer the COBOL (initial) VALUE clause
	 */
	public String getValueClause() {
		return valueClause;
	}

	/**
	 * @see #getValueClause()
	 */
	public void setValueClause(String valueClause) {
		this.valueClause = valueClause;
	}
	
	/**
	 * @see ElementSpec#writeXmlAttributesOn(XMLStreamWriter)
	 * @since 2.4.4
	 */
	@Override
	protected void writeXmlAttributesOn(XMLStreamWriter writer) throws XMLStreamException {
		super.writeXmlAttributesOn(writer);
		
		if (getOccurs() != null) {
			getOccurs().writeXmlAttributesOn(writer);
		}
		if (getValueClause() != null) {
			writer.writeAttribute("value", getValueClause());
		}
		if (getRedefines() != null) {
			writer.writeAttribute("redefines", Integer.toString(getRedefines().getSymbolId()));
		}
	}
	
	/**
	 * Write the children of the receiver in XML format.
	 * This behaviour applies to all StorageSpec subtypes
	 * @since 2.4.4
	 */
	protected void writeXmlChildrenOn(XMLStreamWriter writer, int indentLevel) throws XMLStreamException {
		for (Iterator<ElementSpec> i=getChildren().iterator(); i.hasNext();) {
			ElementSpec spec = i.next();
			spec.writeXmlOn(writer, indentLevel);
		}
	}

	/**
	 * @see ElementSpec#parseXmlFrom(RecordSpecTreeXMLParser)
	 * @since 2.4.4
	 */
	@Override
	protected void parseXmlFrom(RecordSpecTreeXMLParser parser) throws XMLStreamException {
		parseXmlAttributesFrom(parser);
		parser.addSpecById(this);
		parseXmlChildrenFrom(parser);
	}
	
	/**
	 * @see ElementSpec#parseXmlAttributesFrom(RecordSpecTreeXMLParser)
	 * @since 2.4.4
	 */
	protected void parseXmlAttributesFrom(RecordSpecTreeXMLParser parser) throws XMLStreamException {
		super.parseXmlAttributesFrom(parser);
		
		Occurs occurs = Occurs.parseXmlFrom(parser);
		if (occurs instanceof OccursDependingOn) {
			OccursDependingOn odo = (OccursDependingOn)occurs;
			// configure any parent GroupSpec as containing and ODO
			RecordSpecTreeBuilder.configureOdoEnclosures(this);
			// let the target/object of the ODO clause know that it is one
			odo.getOdoObject().setContainsOdoObject(true);
		}
		setOccurs(occurs);
		
		setValueClause(parser.getAttributeValue("value"));
		int redefinesId = parser.getAttributeIntValue("redefines", -1);
		if (redefinesId >= 0) {
			setRedefines(parser.getStorageSpecForId(redefinesId));
		}
	}
	
	/**
	 * Parse children of StorageSpec (and subtypes).
	 * This includes nested ElementSpec (subtypes) as well as source lines.
	 */
	protected void parseXmlChildrenFrom(RecordSpecTreeXMLParser parser) throws XMLStreamException {
		XMLStreamReader reader = parser.getReader();
		String ename;
		StringBuffer sb = new StringBuffer(getSourceLinesAsString());  // add to any existing source
		boolean endElement = false;
		while (!endElement && reader.hasNext()) {
			switch (reader.next()) {
			case XMLStreamConstants.START_ELEMENT:
				ename = reader.getLocalName();
				ElementSpec child;
				if (ename.equalsIgnoreCase(ConditionSpec.XML_ELEMENT_NAME)) {
					child = new ConditionSpec(null);
				} else if (ename.equalsIgnoreCase(FieldSpec.XML_ELEMENT_NAME)) {
					child = new FieldSpec();
				} else if (ename.equalsIgnoreCase(GroupSpec.XML_ELEMENT_NAME)) {
					child = new GroupSpec();
				} else if (ename.equalsIgnoreCase(RenameSpec.XML_ELEMENT_NAME)) {
					child = new RenameSpec(null);
				} else if (ename.equalsIgnoreCase(RenameRangeSpec.XML_ELEMENT_NAME)) {
					child = new RenameRangeSpec(null, null);
				} else {
					throw new XMLStreamException("unrecognized element: "+ename, reader.getLocation());
				}
				addChild(child);
				child.setParent(this);
				child.parseXmlFrom(parser);
				break;
			case XMLStreamConstants.END_ELEMENT:
				endElement = true;
				break;
			case XMLStreamConstants.CHARACTERS:
				sb.append(new String(reader.getTextCharacters(),
										reader.getTextStart(),
										reader.getTextLength()));
				break;
			case XMLStreamConstants.CDATA:
				if (reader.hasText()) {
					sb.append(reader.getText());
				}
				break;
				
			}
		}
		setSourceLines(sb.toString());
	}
}
